# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementations of quantum gates.

The :class:`.GateImplementation` subclasses in this subpackage construct :class:`.TimeBox` instances to
implement specific native gates, using the calibration data that the class has been initialized with.
Each GateImplementation instance encapsulates the calibration data for a specific implementation of a specific
native gate acting on a specific locus.

Several different implementations and calibration schemes can be supported for a given gate,
each represented by its own GateImplementation subclass.
Likewise, a single GateImplementation subclass can be sometimes used to implement several different gates
through different calibration data.
"""

from dataclasses import replace

from iqm.pulse.gate_implementation import GateImplementation
from iqm.pulse.gates.barrier import Barrier
from iqm.pulse.gates.conditional import (
    CCPRX_Composite,
    CCPRX_Composite_DRAGCosineRiseFall,
    CCPRX_Composite_DRAGGaussian,
)
from iqm.pulse.gates.cz import (
    CZ_CRF,
    CZ_CRF_ACStarkCRF,
    CZ_GaussianSmoothedSquare,
    CZ_Slepian,
    CZ_Slepian_ACStarkCRF,
    CZ_Slepian_CRF,
    CZ_TruncatedGaussianSmoothedSquare,
    FluxPulseGate_CRF_CRF,
    FluxPulseGate_TGSS_CRF,
)
from iqm.pulse.gates.default_gates import _quantum_ops_library
from iqm.pulse.gates.delay import Delay
from iqm.pulse.gates.flux_multiplexer import FluxMultiplexer_SampleLinear
from iqm.pulse.gates.measure import Measure_Constant, Measure_Constant_Qnd, Shelved_Measure_Constant
from iqm.pulse.gates.move import MOVE_CRF_CRF, MOVE_SLEPIAN_CRF, MOVE_TGSS_CRF
from iqm.pulse.gates.prx import (
    Constant_PRX_with_smooth_rise_fall,
    PRX_DRAGCosineRiseFall,
    PRX_DRAGCosineRiseFallSX,
    PRX_DRAGGaussian,
    PRX_DRAGGaussianSX,
    PRX_FastDrag,
    PRX_FastDragSX,
    PRX_HdDrag,
    PRX_HdDragSX,
    PRX_ModulatedDRAGCosineRiseFall,
    get_unitary_prx,
)
from iqm.pulse.gates.reset import Reset_Conditional, Reset_Wait
from iqm.pulse.gates.rz import (
    RZ_ACStarkShift_CosineRiseFall,
    RZ_ACStarkShift_smoothConstant,
    RZ_PRX_Composite,
    RZ_Virtual,
    get_unitary_rz,
)
from iqm.pulse.gates.sx import SXGate
from iqm.pulse.gates.u import UGate, get_unitary_u
from iqm.pulse.quantum_ops import QuantumOp, QuantumOpTable

_exposed_implementations: dict[str, type[GateImplementation]] = {
    cls.__name__: cls  # type: ignore[misc]
    for cls in (
        Barrier,
        Constant_PRX_with_smooth_rise_fall,
        PRX_DRAGGaussian,
        PRX_DRAGCosineRiseFall,
        PRX_DRAGGaussianSX,
        PRX_DRAGCosineRiseFallSX,
        PRX_FastDragSX,
        PRX_FastDrag,
        PRX_HdDrag,
        PRX_HdDragSX,
        SXGate,
        UGate,
        RZ_PRX_Composite,
        RZ_Virtual,
        CZ_CRF_ACStarkCRF,
        CZ_Slepian_ACStarkCRF,
        CZ_GaussianSmoothedSquare,
        CZ_Slepian,
        CZ_Slepian_CRF,
        CZ_CRF,
        CZ_TruncatedGaussianSmoothedSquare,
        FluxPulseGate_TGSS_CRF,
        FluxPulseGate_CRF_CRF,
        Measure_Constant,
        Measure_Constant_Qnd,
        Shelved_Measure_Constant,
        PRX_ModulatedDRAGCosineRiseFall,
        MOVE_CRF_CRF,
        MOVE_SLEPIAN_CRF,
        MOVE_TGSS_CRF,
        RZ_ACStarkShift_CosineRiseFall,
        RZ_ACStarkShift_smoothConstant,
        CCPRX_Composite,
        CCPRX_Composite_DRAGCosineRiseFall,
        CCPRX_Composite_DRAGGaussian,
        Reset_Conditional,
    )
}
"""These GateImplementations can be referred to in the configuration YAML."""


def get_implementation_class(class_name: str) -> type[GateImplementation] | None:
    """Get gate implementation class by class name."""
    return _exposed_implementations.get(class_name, None)


def expose_implementation(implementation: type[GateImplementation], overwrite: bool = False) -> None:
    """Add the given gate implementation to the list of known implementations.

    Args:
        implementation: GateImplementation to add so that it can be found with :func:`.get_implementation_class`.
        overwrite: If True, does not raise an error if implementation already exists.

    """
    name = implementation.__name__
    if name in _exposed_implementations:
        if not overwrite and _exposed_implementations[name] is not implementation:
            raise ValueError(f"GateImplementation '{name}' has already been defined.")
    _exposed_implementations[name] = implementation


def _compare_operations(op1: QuantumOp, op2: QuantumOp) -> bool:
    """Compares two QuantumOp instances. Operations are different only if certain parameters do not match.
    (everything else except the implementations information which the user is allowed to modify and
    :attr:`QuatumOp.unitary` the equality of which cannot be validated as it is a free-form function).

    Args:
        op1: First QuantumOp instance
        op2: Second QuantumOp instance

    Returns:
        True if the operations have identical parameters, False otherwise

    """
    IGNORED_FIELDS = {"implementations", "defaults_for_locus", "unitary"}
    op1_dict = vars(op1)
    op2_dict = vars(op2)
    return all(
        (op1_dict[field] == op2_dict[field] or (field == "params" and tuple(op1_dict[field]) == tuple(op2_dict[field])))
        for field in op1_dict
        if field not in IGNORED_FIELDS
    )


def _validate_operation(
    new_op: QuantumOp,
    gate_name: str,
    operations: QuantumOpTable,
    overwrite: bool = False,
) -> QuantumOp:
    """Validate new operation against existing operations and set unitary if needed.

    Args:
        new_op: Operation to validate
        gate_name: Name of the gate
        operations: Dictionary containing existing operations
        overwrite: Whether to allow overwriting existing operations

    Returns:
        Validated QuantumOp with unitary set if needed

    Raises:
        ValueError: If operation exists with different parameters and overwrite is False

    """
    if not overwrite and gate_name in operations:
        old = operations.get(gate_name)
        same = _compare_operations(new_op, old)
        if not same:
            raise ValueError(f"{gate_name} already registered with different parameters")

    if gate_name in _quantum_ops_library:
        default = _quantum_ops_library.get(gate_name)
        same = _compare_operations(new_op, default)
        if not same:
            raise ValueError(f"{gate_name} conflicts with a canonical operation in iqm-pulse")

    if new_op.unitary is None:
        if gate_name in operations and operations[gate_name].unitary is not None:
            unitary = operations[gate_name].unitary
        elif gate_name in _quantum_ops_library and _quantum_ops_library[gate_name].unitary is not None:
            unitary = default.unitary
        else:
            unitary = None
        new_op = replace(new_op, unitary=unitary)

    return new_op


def _register_gate(
    operations: QuantumOpTable,
    gate_name: str,
    impl_class: type[GateImplementation],
    quantum_op_specs: QuantumOp | dict | None = None,
) -> QuantumOp:
    """Create the quantum operation for the new gate.

    Args:
        operations: Known operations, mapping gate names to QuantumOps
        gate_name: Name of the gate to register
        impl_class: The implementation class
        quantum_op_specs: The quantum operation specifications

    Returns:
        The instance of the new quantum operation

    """
    if isinstance(quantum_op_specs, QuantumOp):
        new_op = quantum_op_specs
    elif quantum_op_specs is None and gate_name in operations:
        new_op = operations[gate_name]
    else:
        new_kwargs = {
            "name": gate_name,
            "arity": 1,
            "params": tuple(),
            "implementations": {},
            "symmetric": impl_class.symmetric,
            "factorizable": False,
        }
        if quantum_op_specs:
            new_kwargs |= quantum_op_specs
            new_kwargs["params"] = tuple(new_kwargs.get("params", ()))

        new_op = QuantumOp(**new_kwargs)

    return new_op


def _add_implementation(
    operations: dict[str, QuantumOp],
    new_op: QuantumOp,
    impl_name: str,
    impl_class: type[GateImplementation],
    set_as_default: bool = False,
    overwrite: bool = False,
) -> None:
    """Register a new implementation for an existing quantum operation.

    Args:
        operations: Table of existing quantum operations
        new_op: The new quantum operation
        impl_name: The name for the implementation that is added
        impl_class: The GateImplementation class corresponding to the new implementation
        set_as_default: Whether to set as default implementation
        overwrite: If True, allows replacing existing implementation

    Returns:
        new_op: QuantumOp with the new implementation

    """
    new_op.implementations[impl_name] = impl_class
    _validate_implementation(operations, new_op, impl_name, impl_class)
    if set_as_default and len(new_op.implementations) >= 1:
        new_op.set_default_implementation(impl_name)
    if not get_implementation_class(impl_class.__name__):
        expose_implementation(impl_class, overwrite)

    return new_op


def _validate_implementation(
    operations: QuantumOpTable,
    new_op: QuantumOp,
    impl_name: str,
    impl_class: type[GateImplementation],
) -> None:
    """Validate new implementation against existing implementations.

    Args:
        operations: Table of quantum operations
        new_op: Operation whose implementation we want to validate
        impl_name: Name of the new implementation
        impl_class: The GateImplementation class corresponding to the new
        implementation

    Raises:
        ValueError: If there is already an implementation with the same name, but
        corresponds to a different implementation class.

    """
    if new_op.name in _quantum_ops_library:
        default_op = _quantum_ops_library[new_op.name]
        for default_impl_name, default_impl_cls in default_op.implementations.items():
            if impl_name == default_impl_name:
                default_impl_cls_name = (
                    default_impl_cls if isinstance(default_impl_cls, str) else default_impl_cls.__name__
                )
                if impl_class.__name__ != default_impl_cls_name:
                    raise ValueError(
                        f"The implementation '{default_impl_name}' already exists in default implementations with '{default_impl_cls.__name__}' as corresponding GateImplementation class."  # noqa: E501
                    )

    if new_op.name in operations:
        existing_op = operations[new_op.name]
        for existing_impl_name, existing_impl_cls in existing_op.implementations.items():
            if impl_name == existing_impl_name:
                existing_impl_cls_name = (
                    existing_impl_cls if isinstance(existing_impl_cls, str) else existing_impl_cls.__name__
                )  # noqa: E501
                if impl_class.__name__ != existing_impl_cls_name:
                    raise ValueError(
                        f"The implementation '{existing_impl_name}' already exists with '{existing_impl_cls.__name__}' as corresponding GateImplementation class."  # noqa: E501
                    )


def register_implementation(
    operations: dict[str, QuantumOp],
    gate_name: str,
    impl_name: str,
    impl_class: type[GateImplementation],
    set_as_default: bool = False,
    overwrite: bool = False,
    quantum_op_specs: QuantumOp | dict | None = None,
) -> None:
    """Register a new gate implementation, and a new gate (operation) if needed.

    Args:
        operations: Known operations, mapping gate names to QuantumOps
        gate_name: The gate name to register
        impl_name: The name for this implementation
        impl_class: The implementation class
        set_as_default: Whether to set as default implementation
        overwrite: If True, allows replacing existing operation/implementation
        quantum_op_specs: Specs for creating new quantum op if needed

    Raises:
        ValueError: If operation/implementation exists and overwrite=False

    """
    new_op = _register_gate(operations, gate_name, impl_class, quantum_op_specs)

    new_op = _validate_operation(new_op, gate_name, operations, overwrite)

    _add_implementation(operations, new_op, impl_name, impl_class, set_as_default, overwrite)

    operations[gate_name] = new_op



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Single-qubit PRX gate.

The phased x rotation (PRX) gate is defined as

.. math::
   R_\phi(\theta) = e^{-i (X \cos \phi + Y \sin \phi) \: \theta/2}
    = R_z(\phi) R_x(\theta) R_z^\dagger(\phi),

where the rotation angle :math:`\theta` and the phase angle :math:`\phi` are in radians.

It rotates the qubit state around an axis that lies in the XY plane of the Bloch sphere.
"""

from __future__ import annotations

from abc import abstractmethod
import copy
from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np

from exa.common.data.parameter import Parameter, Setting
from iqm.pulse.gate_implementation import (
    SINGLE_COMPONENTS_WITH_DRIVE_LOCUS_MAPPING,
    CustomIQWaveforms,
    GateImplementation,
    Locus,
    OILCalibrationData,
    SinglePulseGate,
    get_waveform_parameters,
)
from iqm.pulse.gates.enums import XYGate
from iqm.pulse.playlist.fast_drag import FastDragI, FastDragQ
from iqm.pulse.playlist.hd_drag import HdDragI, HdDragQ
from iqm.pulse.playlist.instructions import Block, IQPulse
from iqm.pulse.playlist.schedule import TOLERANCE, Schedule
from iqm.pulse.playlist.waveforms import (
    Constant,
    CosineFall,
    CosineRise,
    CosineRiseFall,
    TruncatedGaussian,
    Waveform,
)
from iqm.pulse.playlist.waveforms import CosineRiseFallDerivative as CosineRiseFallD
from iqm.pulse.playlist.waveforms import TruncatedGaussianDerivative as TruncatedGaussianD
from iqm.pulse.utils import normalize_angle, phase_transformation

if TYPE_CHECKING:  # pragma: no cover
    from iqm.pulse.builder import ScheduleBuilder
    from iqm.pulse.quantum_ops import QuantumOp
    from iqm.pulse.timebox import TimeBox


@lru_cache
def get_unitary_prx(angle: float, phase: float) -> np.ndarray:
    """Unitary for a PRX gate.

    Args:
        angle: rotation angle (in rad)
        phase: phase angle (in rad)

    Returns:
        2x2 unitary representing ``prx(angle, phase)``.

    """
    return np.array(
        [
            [np.cos(angle / 2), -1j * np.exp(-1j * phase) * np.sin(angle / 2)],
            [-1j * np.exp(1j * phase) * np.sin(angle / 2), np.cos(angle / 2)],
        ]
    )


class PrxGateImplementation(GateImplementation):
    """ABC for different implementations of the PRX gate."""

    def __init__(
        self, parent: QuantumOp, name: str, locus: Locus, calibration_data: OILCalibrationData, builder: ScheduleBuilder
    ) -> None:
        super().__init__(parent, name, locus, calibration_data, builder)
        self._cliffords: dict[XYGate, TimeBox] = {}

    @abstractmethod
    def _call(self, angle: float, phase: float) -> TimeBox:
        """Phased X rotation gate.

        Args:
            angle: Rotation angle in radians.
            phase: Phase angle in radians.

        Returns:
            Boxed instruction schedule implementing the phased X rotation gate.

        """
        raise NotImplementedError

    def rx(self, angle: float) -> TimeBox:
        """X rotation gate.

        Args:
            angle: rotation angle (in radians)

        Returns:
            boxed instruction schedule implementing the x rotation gate

        """
        box = self(angle=angle, phase=0)
        box.label = f"Rx on {self.locus[0]}"
        return box

    def ry(self, angle: float) -> TimeBox:
        """Y rotation gate.

        Args:
            angle: rotation angle (in radians)

        Returns:
            boxed instruction schedule implementing the y rotation gate

        """
        box = self(angle=angle, phase=np.pi / 2)
        box.label = f"Ry on {self.locus[0]}"
        return box

    def clifford(self, xy_gate: XYGate) -> TimeBox:
        """One-qubit XY Clifford gates.

        Args:
            xy_gate: Clifford gate

        Returns:
            boxed instruction schedule implementing the requested Clifford gate

        """
        if gate := self._cliffords.get(xy_gate):
            return gate

        # cache Cliffords on first use
        match xy_gate:
            case XYGate.IDENTITY:
                # TODO could be also implemented with a zero-duration Wait, which would be more accurate.
                gate = self.rx(0.0)
            case XYGate.X_90:
                gate = self.rx(np.pi / 2)
            case XYGate.X_180:
                gate = self.rx(np.pi)
            case XYGate.X_M90:
                gate = self.rx(-np.pi / 2)
            case XYGate.Y_90:
                gate = self.ry(np.pi / 2)
            case XYGate.Y_180:
                gate = self.ry(np.pi)
            case XYGate.Y_M90:
                gate = self.ry(-np.pi / 2)
            case _:
                raise ValueError(f"Unknown XYGate: {xy_gate}")
        gate.label = f"{xy_gate.name} on {self.locus[0]}"
        self._cliffords[xy_gate] = gate
        return gate

    @classmethod
    def get_locus_mapping_name(cls, operation_name: str, implementation_name: str) -> str:
        return SINGLE_COMPONENTS_WITH_DRIVE_LOCUS_MAPPING


class PRX_SinglePulse_GateImplementation(SinglePulseGate, PrxGateImplementation):
    r"""Base class for PRX gates implemented using a single IQ pulse.

    This class implements phased x rotation gates on a specific qubit using an :class:`.IQPulse`
    instance, derived from the pulse calibration data provided at construction by
    the static method :meth:`_single_iq_pulse`.
    The pulse is used to resonantly drive the qubit to effect the required rotation.

    The pulse calibration data consists of the parameters of an :math:`x_\pi` :class:`.IQPulse` only.
    It is assumed that

    * the transfer function from the AWG to the qubit is linear, i.e.,
      other rotation angles can be obtained by linearly scaling the pulse amplitude, and
    * other phase angles can be obtained by adjusting the IQ modulation phase.

    The generated pulses all have the same time duration, also for identity rotations. In the special case of the
    duration being zero, the gate implementation will apply a ``Block(0)`` instruction to the qubit's drive channel.
    """

    def _call(self, angle: float, phase: float = 0.0) -> TimeBox:
        scale, new_phase = _normalize_params(angle, phase)
        pulse = self.pulse.copy(
            scale_i=scale * self.pulse.scale_i,
            scale_q=scale * self.pulse.scale_q,
            phase=self.pulse.phase + new_phase,
        )
        if self.pulse.duration > TOLERANCE:
            timebox = self.to_timebox(Schedule({self.channel: [pulse]}))
        else:
            timebox = self.to_timebox(Schedule({self.channel: [Block(0)]}))
        timebox.neighborhood_components[0] = set(self.locus)
        return timebox

    @property
    def iq_pulse(self) -> IQPulse:
        """Alias for ``self.pulse`` for backward compatibility"""
        return self.pulse


class PRX_CustomWaveforms(PRX_SinglePulse_GateImplementation, CustomIQWaveforms):
    """Base class for PRX gates implemented using a single IQ pulse and hot-swappable waveforms."""

    root_parameters: dict[str, Parameter | Setting] = {
        "duration": Parameter("", "pi pulse duration", "s"),
        "amplitude_i": Parameter("", "pi pulse I channel amplitude", ""),
        "amplitude_q": Parameter("", "pi pulse Q channel amplitude", ""),
    }

    @classmethod
    def _get_pulse(  # type: ignore[override]
        cls,
        *,
        amplitude_i: float,
        amplitude_q: float,
        n_samples: int,
        **rest_of_calibration_data,
    ) -> IQPulse:
        """Builds an x_pi pulse out of the calibration data."""
        if cls.dependent_waves:
            wave_i = cls.wave_i(n_samples=n_samples, **rest_of_calibration_data)
            wave_q = cls.wave_q(n_samples=n_samples, **rest_of_calibration_data)
        else:
            wave_i = cls.wave_i(n_samples=n_samples, **rest_of_calibration_data["i"])
            wave_q = cls.wave_q(n_samples=n_samples, **rest_of_calibration_data["q"])
        return IQPulse(
            n_samples,
            wave_i=wave_i,
            wave_q=wave_q,
            scale_i=amplitude_i,
            scale_q=amplitude_q,
            phase=0,
            phase_increment=0,
        )


class PRX_DRAGGaussian(PRX_CustomWaveforms, wave_i=TruncatedGaussian, wave_q=TruncatedGaussianD):  # type:ignore[call-arg]
    """PRX gate, DRAG / TruncatedGaussian IQ pulse implementation.

    See :class:`.PRX_CustomWaveforms`.
    """


class PRX_DRAGCosineRiseFall(PRX_CustomWaveforms, wave_i=CosineRiseFall, wave_q=CosineRiseFallD):  # type:ignore[call-arg]
    """PRX gate, DRAG / CosineRiseFall IQ pulse implementation.

    See :class:`.PRX_CustomWaveforms`.
    """

    excluded_parameters = ["rise_time"]

    @classmethod
    def _get_pulse(cls, **kwargs) -> IQPulse:
        kwargs["rise_time"] = kwargs["full_width"] / 2  # just a cosine, no flat part
        return super()._get_pulse(**kwargs)


class PRX_CustomWaveformsSX(PRX_SinglePulse_GateImplementation, CustomIQWaveforms):
    r"""Base class for PRX gates implemented using SX gate, hot-swappable waveforms and phase manipulation.

    The schedule used to implement the PRX gate depends on the arguments:
        1. If the rotation angle :math:`\theta = \pi/2`, the timebox will consist of just the SX IQ pulse, with phase.
        2. If the rotation angle :math:`\theta = 0.0`, the timebox will consist of a single zero-amplitude pulse.
        3. If not, the timebox will consist of two IQ pulses, with phase.

    The formula for the PRX gate implemented using SX gates and z rotations is

    .. math::
       R_\phi(\theta) = R_z(\phi-\pi/2) \: \text{SX} \: R_z(\pi-\theta) \: \text{SX} \: R_z(-\phi-\pi/2).

    The fusing of z rotations to IQPulses is done inside the :meth:`_call` method.

    All parameters in the pulse here is referring to the state of the qubits.
    """

    root_parameters: dict[str, Parameter | Setting] = {
        "duration": Parameter("", "pi pulse duration", "s"),
        "amplitude_i": Parameter("", "pi pulse I channel amplitude", ""),
        "amplitude_q": Parameter("", "pi pulse Q channel amplitude", ""),
        "rz_before": Parameter("", "RZ before IQ pulse", "rad"),
        "rz_after": Parameter("", "RZ after IQ pulse", "rad"),
    }

    def _call(self, angle: float, phase: float = 0) -> TimeBox:  # type: ignore[override]
        """Convert pulses into timebox, via extra Z rotations.

        There are exceptions while using 0, pi/2 and pi rotation in angle, for calibration reason. The duration of the
        timebox can be different.
        """
        new_angle, new_phase = _normalize_params(angle, phase)
        new_angle = new_angle * np.pi
        if self.pulse.duration < TOLERANCE:
            timebox = self.to_timebox(Schedule({self.channel: [Block(0)]}))
        elif np.isclose(new_angle, 0.0, atol=1e-8):
            # Play zero-amplitude pulse similarly to PRX_CustomWaveforms. The phase increment must be set to zero
            # so that the pulse induces no Z-rotation
            pulse = self.pulse.copy(scale_i=0.0, scale_q=0.0, phase=new_phase, phase_increment=0.0)
            timebox = self.to_timebox(Schedule({self.channel: [pulse]}))
        elif np.isclose(new_angle, np.pi / 2, atol=1e-8):
            pulse = self.pulse.copy(
                phase=self.pulse.phase + new_phase,
            )
            timebox = self.to_timebox(Schedule({self.channel: [pulse]}))
        elif np.isclose(new_angle, np.pi, atol=1e-8):
            pulse = self.pulse.copy(
                phase=self.pulse.phase + new_phase,
            )
            timebox = self.to_timebox(Schedule({self.channel: [pulse, pulse]}))
        else:
            rz_a = -np.pi / 2 - new_phase
            rz_b = np.pi - new_angle
            rz_c = -np.pi / 2 + new_phase
            phase_1, phase_increment_1 = phase_transformation(rz_a, 0)
            phase_2, phase_increment_2 = phase_transformation(rz_b, rz_c)
            pulse_1 = self.pulse.copy(
                phase=normalize_angle(self.pulse.phase + phase_1),
                phase_increment=normalize_angle(self.pulse.phase_increment + phase_increment_1),
            )
            pulse_2 = self.pulse.copy(
                phase=normalize_angle(self.pulse.phase + phase_2),
                phase_increment=normalize_angle(self.pulse.phase_increment + phase_increment_2),
            )
            timebox = self.to_timebox(Schedule({self.channel: [pulse_1, pulse_2]}))

        timebox.neighborhood_components[0] = set(self.locus)
        return timebox

    @classmethod
    def _get_pulse(  # type: ignore[override]
        cls,
        *,
        amplitude_i: float,
        amplitude_q: float,
        n_samples: int,
        **rest_of_calibration_data,
    ) -> IQPulse:
        """Builds a single sqrt(X) pulse from the calibration data."""
        rz_before = rest_of_calibration_data.pop("rz_before", 0)
        rz_after = rest_of_calibration_data.pop("rz_after", 0)
        phase, phase_increment = phase_transformation(rz_before, rz_after)

        if cls.dependent_waves:
            wave_i = cls.wave_i(n_samples=n_samples, **rest_of_calibration_data)
            wave_q = cls.wave_q(n_samples=n_samples, **rest_of_calibration_data)
        else:
            wave_i = cls.wave_i(n_samples=n_samples, **rest_of_calibration_data["i"])
            wave_q = cls.wave_q(n_samples=n_samples, **rest_of_calibration_data["q"])
        return IQPulse(
            n_samples,
            wave_i=wave_i,
            wave_q=wave_q,
            scale_i=amplitude_i,
            scale_q=amplitude_q,
            phase=phase,
            phase_increment=phase_increment,
        )


class PRX_DRAGGaussianSX(PRX_CustomWaveformsSX, wave_i=TruncatedGaussian, wave_q=TruncatedGaussianD):  # type:ignore[call-arg]
    """PRX gate, DRAG / Gaussian IQ pulse with VZ implementation.

    See :class:`.PRX_CustomWaveformsVZ`.
    """


class PRX_DRAGCosineRiseFallSX(PRX_CustomWaveformsSX, wave_i=CosineRiseFall, wave_q=CosineRiseFallD):  # type:ignore[call-arg]
    """PRX gate, DRAG / CosineRiseFall IQ pulse with VZ implementation.

    See :class:`.PRX_CustomWaveformsVZ`.
    """

    excluded_parameters = ["rise_time"]

    @classmethod
    def _get_pulse(cls, **kwargs) -> IQPulse:
        kwargs["rise_time"] = kwargs["full_width"] / 2  # just a cosine, no flat part
        return super()._get_pulse(**kwargs)


class PRX_FastDragSX(PRX_CustomWaveformsSX, wave_i=FastDragI, wave_q=FastDragQ):  # type:ignore[call-arg]
    """PRX gate, FAST DRAG IQ pulse with VZ-based SX-implementation.

    See :class:`.PRX_CustomWaveformsSX`.
    """


class PRX_FastDrag(PRX_CustomWaveforms, wave_i=FastDragI, wave_q=FastDragQ):  # type:ignore[call-arg]
    """PRX gate, FAST DRAG IQ pulse based on amplitude scaling.

    See :class:`.PRX_CustomWaveforms`.
    """


class PRX_HdDragSX(PRX_CustomWaveformsSX, wave_i=HdDragI, wave_q=HdDragQ):  # type:ignore[call-arg]
    """PRX gate, HD DRAG IQ pulse with VZ-based SX-implementation.

    See :class:`.PRX_CustomWaveformsSX`.
    """


class PRX_HdDrag(PRX_CustomWaveforms, wave_i=HdDragI, wave_q=HdDragQ):  # type:ignore[call-arg]
    """PRX gate, HD DRAG IQ pulse based on amplitude scaling

    See :class:`.PRX_CustomWaveforms`.
    """


def _normalize_params(angle: float, phase: float) -> tuple[float, float]:
    """Algebraic normalization of ``angle`` to [0, pi] and ``phase`` to (-pi, pi].

    Returns:
        amplitude scaling factor, phase for a pi pulse

    """
    # 1. Normalize angle to (-pi, pi].
    # 2. Use the symmetry PRX(-theta, phi ± pi) = PRX(theta, phi) to move rotation angle in [0, pi]
    # 3. Normalize phase to (-pi, pi].
    # FIXME: ideally, here we want to do normalization only. Step 2 is mostly motivated by some rules set in ZI
    # compiler (see COMP-544 for details), and we shall get rid of it or move somewhere more ZI specific.
    half_turn = np.pi
    angle = normalize_angle(angle)
    if angle < 0:
        angle = -angle
        phase += half_turn
    return angle / half_turn, normalize_angle(phase)


class ABC_Constant_smooth(PrxGateImplementation):
    r"""Base class for creating gates with an arbitrarily long Constant pulses with smooth rise and fall.
    This pulse creates a :'Segment' consisting of three instructions : [rise_waveform, main_waveform, fall_waveform].
    This class is created so that one can use middle waveform as a constant, thus enabling to use arbitrarily
    long pulses, not limited by the awg memory.

    Attributes::
        main_waveform: The middle part of the pulse, which should (but doesn't have to) be a Constant waveform
        rise_waveform: rise part of the pulse
        fall_waveform: fall part of the pulse
        channel: Name of the drive channel on which the AC Stark pulse is played.

    """

    main_waveform: type[Waveform]
    rise_waveform: type[Waveform]
    fall_waveform: type[Waveform]

    def __init__(
        self,
        parent: QuantumOp,
        name: str,
        locus: Locus,
        calibration_data: OILCalibrationData,
        builder: ScheduleBuilder,
    ):
        """Constructs an instance of the AC Stark pulse for the given locus."""
        super().__init__(parent, name, locus, calibration_data, builder)
        drive_channel = builder.get_drive_channel(locus[0])

        params = self.convert_calibration_data(calibration_data, self.parameters, builder.channels[drive_channel])

        params_for_stark = copy.deepcopy(params)
        params_for_risefall = copy.deepcopy(params)

        params_for_stark["n_samples"] = max(
            int(round(params_for_stark["n_samples"] * (1 - 2 * params_for_stark["rise_time"]), 0)), 0
        )
        self.main_waveform = self._main_pulse(**params_for_stark)

        params_for_risefall["n_samples"] = int(
            round(params_for_risefall["n_samples"] * params_for_risefall["rise_time"], 0)
        )

        self.fall_waveform = self._fall_pulse(**params_for_risefall)
        self.rise_waveform = self._rise_pulse(**params_for_risefall)

        self.channel = drive_channel
        self.special_implementation = True  # DEBUG

    def __init_subclass__(
        cls,
        /,
        fall_waveform: type[Waveform],
        rise_waveform: type[Waveform],
        main_waveform: type[Waveform],
    ):
        """Store the Waveform types used by this subclass, and their parameters."""
        cls.main_waveform = main_waveform
        cls.rise_waveform = rise_waveform
        cls.fall_waveform = fall_waveform

        cls.symmetric = True
        parameters = get_waveform_parameters(main_waveform)
        parameters["amplitude_i"] = Parameter("", "amplitude_i", "")
        parameters["amplitude_q"] = Parameter("", "amplitude_q", "")

        parameters["rise_time"] = Parameter("", "gate rise time", "s")

        cls.parameters = {
            "duration": Parameter("", "Gate duration", "s"),
        } | parameters

    def _call(self, angle: float, phase: float = 0.0) -> TimeBox:
        scale, new_phase = _normalize_params(angle, phase)

        pulse_rise = self.rise_waveform.copy(
            scale_i=scale * self.rise_waveform.scale_i,
            scale_q=scale * self.rise_waveform.scale_q,
            phase=self.rise_waveform.phase + new_phase,
        )
        pulse_fall = self.fall_waveform.copy(
            scale_i=scale * self.fall_waveform.scale_i,
            scale_q=scale * self.fall_waveform.scale_q,
            phase=self.fall_waveform.phase + new_phase,
        )
        stark_pulse = self.main_waveform.copy(
            scale_i=scale * self.main_waveform.scale_i,
            scale_q=scale * self.main_waveform.scale_q,
            phase=self.main_waveform.phase + new_phase,
        )

        return self.to_timebox(
            Schedule(
                {self.channel: [pulse_rise, stark_pulse, pulse_fall]},
            )
        )

    @classmethod
    def _main_pulse(
        cls,
        *,
        n_samples: int,
        amplitude_i: float,
        amplitude_q: float,
        phase: float = 0.0,
        **kwargs,
    ) -> IQPulse:
        """Returns the main part pulse. Waveform is the same for both I and Q channels"""
        wave = cls.main_waveform(n_samples=n_samples)

        return IQPulse(
            n_samples,
            wave_i=wave,
            wave_q=wave,
            scale_i=amplitude_i,
            scale_q=amplitude_q,
        )

    @classmethod
    def _rise_pulse(
        cls,
        *,
        n_samples: int,
        amplitude_i: float,
        amplitude_q: float,
        **kwargs,
    ) -> IQPulse:
        """Returns a rise pulse."""
        wave_i = cls.rise_waveform(n_samples=n_samples)
        wave_q = cls.rise_waveform(n_samples=n_samples)

        return IQPulse(
            n_samples,
            wave_i=wave_i,
            wave_q=wave_q,
            scale_i=amplitude_i,
            scale_q=amplitude_q,
        )

    @classmethod
    def _fall_pulse(
        cls,
        *,
        n_samples: int,
        amplitude_i: float,
        amplitude_q: float,
        **kwargs,
    ) -> IQPulse:
        """Returns a fall pulse"""
        return IQPulse(
            n_samples,
            wave_i=cls.fall_waveform(n_samples=n_samples),
            wave_q=cls.fall_waveform(n_samples=n_samples),
            scale_i=amplitude_i,
            scale_q=amplitude_q,
        )


class Constant_PRX_with_smooth_rise_fall(
    ABC_Constant_smooth,
    rise_waveform=CosineRise,
    main_waveform=Constant,
    fall_waveform=CosineFall,
):
    """Constant PRX pulse with cosine rise and fall padding.
    Implemented as a 3-instruction Schedule.
    """


class PRX_ModulatedCustomWaveForms(PRX_CustomWaveforms):
    r"""Base class for PRX gates with modulated frequency, hot-swappable waveforms.

    The class takes baseband I and Q waveform as input, and modulates them with frequency in the root_parameters.
    The final pulse shape after modulation is:

    .. math::
        A_I^{\delta}\Omega_I(t)\cos((\omega_d + \delta)t) - A_Q^{\delta}\Omega_Q(t)\sin((\omega_d + \delta)t)

    where :math:`A_I` is `amplitude_i`, :math:`A_Q` is `amplitude_q`, :math:`\Omega` is arbitrary waveform in
    baseband, :math:`\omega_d/2\pi` is the drive frequency and :math:`\delta/2\pi` is the modulated `frequency`.

    """

    root_parameters: dict[str, Parameter | Setting] = {
        "duration": Parameter("", "pi pulse duration", "s"),
        "amplitude_i": Parameter("", "pi pulse amplitude of base band I waveform", ""),
        "amplitude_q": Parameter("", "pi pulse amplitude of base band Q waveform", ""),
        "frequency": Parameter("", "modulated pulse frequency", "Hz"),
    }

    @classmethod
    def _get_pulse(  # type: ignore[override]
        cls, *, amplitude_i: float, amplitude_q: float, n_samples: int, **rest_of_calibration_data
    ) -> IQPulse:
        """Return the IQPulse with modulated arbitrary waveform based on the calibration data."""
        frequency = rest_of_calibration_data.pop("frequency")
        if cls.dependent_waves:
            wave_i = cls.wave_i(n_samples=n_samples, **rest_of_calibration_data)
            wave_q = cls.wave_q(n_samples=n_samples, **rest_of_calibration_data)
        else:
            wave_i = cls.wave_i(n_samples=n_samples, **rest_of_calibration_data["i"])
            wave_q = cls.wave_q(n_samples=n_samples, **rest_of_calibration_data["q"])
        return IQPulse(
            n_samples,
            wave_i=wave_i,
            wave_q=wave_q,
            scale_i=amplitude_i,
            scale_q=amplitude_q,
            modulation_frequency=frequency / n_samples,
        )


class PRX_ModulatedDRAGCosineRiseFall(PRX_ModulatedCustomWaveForms, wave_i=CosineRiseFall, wave_q=CosineRiseFallD):  # type:ignore[call-arg]
    """Modulated PRX pulse with cosine rise fall waveform"""

    excluded_parameters = ["rise_time"]

    @classmethod
    def _get_pulse(cls, **kwargs) -> IQPulse:
        kwargs["rise_time"] = kwargs["full_width"] / 2  # just a cosine, no flat part
        return super()._get_pulse(**kwargs)



#  ********************************************************************************
#  Copyright (c) 2019-2024 IQM Finland Oy.
#  All rights reserved. Confidential and proprietary.
#
#  Distribution or reproduction of any information contained herein
#  is prohibited without IQM Finland Oy’s prior written permission.
#  ********************************************************************************
"""GateImplementation for correcting flux crosstalk for a given set of flux-pulse TimeBoxes"""

from __future__ import annotations

from collections.abc import Iterable
from copy import deepcopy

from iqm.models.playlist.waveforms import Samples
import numpy as np

from exa.common.data.parameter import CollectionType, DataType, Parameter
from exa.common.qcm_data.chip_topology import ChipTopology
from iqm.pulse.gate_implementation import GateImplementation
from iqm.pulse.playlist.instructions import RealPulse, Wait
from iqm.pulse.playlist.schedule import Segment
from iqm.pulse.timebox import TimeBox

TOLERANCE = 1e-10
"""Tolerance for the maximum absolute value in a waveform for considering it to be uniformly zero."""


def _assert_flux_pulse_equality(pulse1: RealPulse, pulse2: RealPulse, assert_waveform: bool = True) -> None:
    error_msg = "FluxMultiplexer requires all multiplexed flux pulses to have equal durations"
    if assert_waveform:
        error_msg += " and waveforms (the amplitude can differ)"
    error = ValueError(error_msg)
    if pulse1.duration != pulse2.duration:
        raise error
    if assert_waveform and pulse1.wave != pulse2.wave:
        raise error


class FluxMultiplexer_SampleLinear(GateImplementation):
    # pylint: disable=anomalous-backslash-in-string, too-many-branches
    r"""Linear flux pulse multiplexed (user for correcting flux crosstalk).

    The required calibration data is the flux cross-talk correction matrix, where the element :math:`C_ij` represents
    the correction needed for flux component ``i`` arising from the crosstalk caused by flux component ``j``, so that
    after the corrections, the flux pulse played at ``i`` is :math:`f(t) = A_i w_i(t) + \sum_j C_ij A_j w_j(t)`,
    where :math:`A_j`` is the flux pulse amplitude for ``j`` and :math:`w_j(t)` the (normalized) waveform.

    The flux crosstalk correction matrix is given in a sparse form via two calibration parameters (we do not support
    dict- or xarray-valued Parameters yet...). Parameter ``matrix_index`` lists the relevant (non-zero) elements of the
    matrix as a flat ``np.array`` of strings of the form ``<flux component i>__<flux component j>``. Parameter
    ``matrix_elements`` lists the corresponding matrix values :math:`C_ij` (the lengths of these arrays must match).

    TODO: this is for now an experimental R&D implementation, and everything here is subject to change still
    """

    parameters = {
        "matrix_index": Parameter(
            "matrix_index",
            label="Flux crosstalk correction sparse matrix index",
            data_type=DataType.STRING,
            collection_type=CollectionType.LIST,
        ),
        "matrix_elements": Parameter(
            "matrix_elements",
            label="Flux crosstalk correction sparse matrix elements",
            data_type=DataType.FLOAT,
            collection_type=CollectionType.NDARRAY,
        ),
    }

    def __call__(self, to_be_multiplexed: Iterable[TimeBox] | TimeBox) -> TimeBox:  # noqa: PLR0912
        """TimeBox where flux-crosstalk errors have been corrected using the crosstalk matrix for a set of
        flux-pulse TimeBoxes.

        The following limitations apply:

        - All the TimeBoxes must be either pure flux-pulses (real pulse on a flux channel) or TimeBoxes created by a
        CZ or MOVE gate ``__call__`` method (real pulse on a flux channel, VirtualRZs on drive channels).
        - All the flux pulses must have the same duration.
        - Any VirtualRZ angles will not be adjusted after doing the flux pulse corrections, this must for now be done
          outside this gate implementation.
        - The flux crosstalk is now corrected only for flux-capable components. There are ideas how to potentially
        correct them for "slow qubits" as well, but this is not implemented yet.

        Args:
            to_be_multiplexed: TimeBoxes to be corrected for flux-crosstalk.

        Returns:
            Equivalent crosstalk-corrected flux pulses.

        """
        # TODO: reimplement the matrix stuff with e.g. scipy sparse matrices
        # TODO: assert that `to_be_multiplexed` contains only allowed TimeBoxes
        matrix_index = self.calibration_data["matrix_index"]
        matrix_elements = self.calibration_data["matrix_elements"]
        # get all crosstalk flux channels
        crosstalk_components = sorted({c for m_ind in matrix_index for c in m_ind.split("__")})
        crosstalk_channels = [self.builder.get_flux_channel(c) for c in crosstalk_components]
        no_flux = [channel for channel, component in zip(crosstalk_channels, crosstalk_components) if channel is None]
        if no_flux:
            raise ValueError(f"Flux crosstalk matrix components {no_flux} have no flux channels.")
        # schedule the fluxes
        if isinstance(to_be_multiplexed, Iterable):
            to_be_multiplexed = TimeBox.composite(to_be_multiplexed)
        scheduled_fluxes = self.builder.resolve_timebox(to_be_multiplexed, neighborhood=0)
        # find the RealPulse template
        real_pulse_candidate = next(p for _, seg in scheduled_fluxes.items() for p in seg if isinstance(p, RealPulse))
        real_pulse_dict = {
            "duration": real_pulse_candidate.duration,
            "wave": real_pulse_candidate.wave,
        }
        # Fix the waits in the flux channels into "duration"-length tiles.
        # We can do this because a) to_be_multiplexed contains only pure fLux pulses and CZ/MOVE TimeBoxes
        # and b) the durations of all flux pulses must be equal.
        for channel in crosstalk_channels:
            if channel in scheduled_fluxes:
                seg = scheduled_fluxes[channel]
                chopped_seg = []
                for inst in seg:
                    if isinstance(inst, Wait):
                        multiple = int(inst.duration / real_pulse_dict["duration"])
                        chopped_seg.extend(multiple * [Wait(real_pulse_dict["duration"])])
                    else:
                        chopped_seg.append(inst)
                scheduled_fluxes[channel] = Segment(chopped_seg)
        max_flux_depth = next(len(seg) for ch, seg in scheduled_fluxes.items() if ch in crosstalk_channels)
        # pad the relevant flux channels not present in the schedule at all with correct number of wait tiles
        for channel in crosstalk_channels:
            if channel not in scheduled_fluxes:
                scheduled_fluxes[channel] = Segment(max_flux_depth * [Wait(real_pulse_dict["duration"])])
        # go through the schedule, layer by layer, and multiplex the flux pulses in each
        primary_components = [p.split("__")[0] for p in matrix_index]
        corrected_fluxes = deepcopy(scheduled_fluxes)
        for layer in range(max_flux_depth):
            for component in primary_components:
                flux_channel = self.builder.get_flux_channel(component)
                crosstalk_links = [
                    (i, p.split("__")[1]) for i, p in enumerate(matrix_index) if component == p.split("__")[0]
                ]
                primary_pulse = scheduled_fluxes[flux_channel][layer]
                if isinstance(primary_pulse, RealPulse):
                    _assert_flux_pulse_equality(primary_pulse, real_pulse_candidate, assert_waveform=False)
                multiplexed_samples = (
                    primary_pulse.scale * primary_pulse.wave.sample()
                    if isinstance(primary_pulse, RealPulse)
                    else np.zeros(primary_pulse.duration)
                )
                for matrix_idx, corrected_component in crosstalk_links:
                    corrected_channel = self.builder.get_flux_channel(corrected_component)
                    corrected_pulse = scheduled_fluxes[corrected_channel][layer]
                    if isinstance(corrected_pulse, RealPulse):
                        _assert_flux_pulse_equality(corrected_pulse, real_pulse_candidate, assert_waveform=False)
                        multiplexed_samples += (
                            matrix_elements[matrix_idx] * corrected_pulse.scale * corrected_pulse.wave.sample()
                        )
                if max(abs(multiplexed_samples)) < TOLERANCE:
                    multiplexed_pulse = Wait(real_pulse_dict["duration"])
                else:
                    multiplexed_pulse = RealPulse(
                        scale=1.0, duration=real_pulse_dict["duration"], wave=Samples(multiplexed_samples)
                    )
                corrected_fluxes[flux_channel]._instructions[layer] = multiplexed_pulse
        locus_components = {c for c in primary_components if c in self.builder.chip_topology.qubits}
        return TimeBox.atomic(
            corrected_fluxes,
            locus_components=locus_components,
            label=f"Cross-talk corrected flux pulses on {locus_components}",
        )

    @classmethod
    def get_custom_locus_mapping(
        cls, chip_topology: ChipTopology, component_to_channels: dict[str, Iterable[str]]
    ) -> dict[tuple[str, ...] | frozenset[str], tuple[str, ...]] | None:
        """Locus is "global" (the whole QPU) represented by an empty tuple for now."""
        # pylint: disable=unused-argument
        return {tuple(): tuple()}



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Two-qubit MOVE gate.

The MOVE gate is a population exchange operation between a qubit and a resonator,
mediated by a coupler, that has the following properties:

* MOVE is unitary.
* The effect of MOVE is only defined in the invariant
  subspace :math:`S = \text{span}\{|00\rangle, |01\rangle, |10\rangle\}`, where it swaps the populations of the states
  :math:`|01\rangle` and :math:`|10\rangle`. Anything may happen in the orthogonal subspace as long as it is unitary and
  invariant.
* In the subspace where it is defined, MOVE is an involution: :math:`\text{MOVE}_S^2 = I_S`.

Thus MOVE has the following presentation in the subspace :math:`S`:

.. math:: \text{MOVE}_S = |00\rangle \langle 00| + a |10\rangle \langle 01| + a^{-1} |01\rangle \langle 10|,

where :math:`a` is an undefined complex phase. This degree of freedom (in addition to the undefined effect of the gate
in the orthogonal subspace) means there is a continuum of different MOVE gates, all equally valid.
The phase :math:`a` is canceled when the MOVE gate is applied a second time due to the involution property.
"""

from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

from exa.common.data.parameter import Parameter, Setting
from iqm.pulse.gates.cz import FluxPulseGate
from iqm.pulse.playlist.instructions import Block, Instruction, VirtualRZ, Wait
from iqm.pulse.playlist.schedule import Schedule
from iqm.pulse.playlist.waveforms import CosineRiseFall, Slepian, TruncatedGaussianSmoothedSquare
from iqm.pulse.timebox import TimeBox
from iqm.pulse.utils import normalize_angle

if TYPE_CHECKING:  # pragma: no cover
    from iqm.pulse.builder import CircuitOperation, ScheduleBuilder


@dataclass(frozen=True, kw_only=True)
class MoveMarker(Wait):
    """Special annotation instruction to indicate the beginning and ending of MOVE gates.

    The *same instance* of this instruction will be inserted into the qubit drive channel and the
    resonator virtual drive channel, right before the beginning MOVE VirtualRZ instructions,
    to link the channels together (otherwise, there would be nothing explicit in the Schedule
    indicating that there is a MOVE gate happening between the qubit and the resonator).

    Another shared instance will be inserted to the aforementioned channels right before the ending
    MOVE VirtualRZ instruction. The VirtualRZ instructions between the markers on the resonator
    channel will be applied to the qubit instead in a post-compilation pass.
    """

    duration: int = 0
    qubit: str
    resonator: str
    detuning: float


class MOVE_CustomWaveforms(FluxPulseGate):
    """Qubit-resonator MOVE gate using flux pulses on both components.

    This class implements the extra phase bookkeeping logic required to make the MOVE
    gates work as intended. Due to the unknown phase in the MOVE gate definition, the MOVEs
    need to be applied in pairs, i.e. the resonator state is always moved back to the qubit
    it came from. Between this pair of MOVE gates you can apply any number of other two-component
    gates (CZs for example) between the resonator and other qubits. This sequence of
    gates acting on the resonator is called a *MOVE sandwich*. At the end of a sandwich we have
    to apply a local phase correction (z rotation) on the state that was moved back to the qubit.

    The :meth:`__call__` method of this class uses the :class:`.MoveMarker` annotation instruction
    to mark the beginning and end of each MOVE sandwich, in order to enable the calculation of the
    angle of the z rotation to be applied on the moved qubit at the end of the sandwich to
    counteract the phase accumulation during the sandwich relative to the computational frame of
    the qubit.
    The phase accumulation has two sources:

    * Phase due to the frequency detuning between the qubit and the resonator,
      proportional to the time duration the MOVE sandwich.

    * Phase due to the virtual z rotations applied on the resonator as
      gates are applied between it and another qubit, which need to be summed up.
      By convention the resonator VirtualRZ angle of the MOVE implementation itself is currently
      always zero (since only the sum of the resonator and qubit z rotation angles matters for MOVE),
      but we also include it in the sum for completeness.

    The phases are calculated and applied on the qubits using :func:`.apply_move_gate_phase_corrections`.
    """

    root_parameters: dict[str, Parameter | Setting] = {
        "duration": Parameter("", "Gate duration", "s"),
        "rz": {
            "*": Parameter("", "Z rotation angle", "rad"),  # wildcard parameter
        },
        "detuning": Parameter("", "Qubit - resonator detuning", "Hz"),
    }
    """Include the frequency difference between qubit and resonator in the gate parameters for phase tracking."""

    def _call(self) -> TimeBox:
        qubit, resonator = self.locus
        qubit_drive_channel = self.builder.get_drive_channel(qubit)
        resonator_drive_channel = self.builder.get_drive_channel(resonator)
        detuning = self.calibration_data["detuning"]

        # special annotated zero-duration Wait instruction appearing in both drive channels
        marker = MoveMarker(qubit=qubit, resonator=resonator, detuning=detuning)
        marker_box = self.to_timebox(
            Schedule(
                {
                    qubit_drive_channel: [marker],
                    resonator_drive_channel: [marker],
                }
            )
        )
        move_box = super()._call()  # box implementing MOVE
        # first the marker, then the MOVE gate
        return TimeBox.composite([marker_box, move_box], label=move_box.label)


class MOVE_CRF_CRF(MOVE_CustomWaveforms, coupler_wave=CosineRiseFall, qubit_wave=CosineRiseFall):
    # type: ignore[call-arg]
    """Qubit-resonator MOVE gate using the CRF waveform for the coupler and the qubit flux pulse."""


class MOVE_SLEPIAN_CRF(MOVE_CustomWaveforms, coupler_wave=Slepian, qubit_wave=CosineRiseFall):
    # type: ignore[call-arg]
    """Qubit-resonator MOVE gate using the Slepian waveform for the coupler flux pulse and the
    CRF waveform for the qubit flux pulse.
    """


class MOVE_TGSS_CRF(MOVE_CustomWaveforms, coupler_wave=TruncatedGaussianSmoothedSquare, qubit_wave=CosineRiseFall):
    # type: ignore[call-arg]
    """Qubit-resonator MOVE gate using the TGSS waveform for the coupler flux pulse and the
    CRF waveform for the qubit flux pulse.
    """


def apply_move_gate_phase_corrections(  # noqa: PLR0915
    schedule: Schedule,
    builder: ScheduleBuilder,
    apply_detuning_corrections: bool = True,
) -> Schedule:
    """Schedule-level pass applying resonator-related phase corrections in MOVE sandwiches to the moved qubit.

    .. note:: Assumes the MOVE gate implementation is based on :class:`.MOVE_CustomWaveforms`.

    Processes all the MOVE sandwiches in ``schedule``, summing up the :class:`.VirtualRZ` instructions
    on the resonator virtual drive channels, adding the phase difference resulting from
    qubit-resonator detuning to the total, and applying it on the qubit at the end of each sandwich.

    Args:
        schedule: instruction schedule to process
        builder: schedule builder that was used to build ``schedule``
        apply_detuning_corrections: if True, also apply detuning phase corrections
    Returns:
        copy of ``schedule`` with the phase corrections applied

    """

    def handle_move_rz(inst: Instruction, component: str) -> float:
        """Handle a RZ from the MOVE gate, acting on the qubit or resonator drive channel.

        Checks that ``inst`` is of the type we expect, and returns its ``phase_increment``.
        """
        if not isinstance(inst, VirtualRZ):
            raise ValueError(f"{type(inst)} following MoveMarker on {component}, should not happen.")
        return inst.phase_increment  # by convention this is currently zero
        # there is no duration-based phase shift for the RZ since it's a part of the MOVE gate
        # and all local phase shifts are already included in the RZs

    def find_phase_corrections() -> dict[MoveMarker, float]:
        """Loop over resonator virtual drive channels to find the phase corrections for
        the MOVE sandwiches.

        Returns:
            mapping from ``MoveMarker`` marking the end of a MOVE sandwich to the total phase
            correction for that sandwich, in radians
            (the same MoveMarker instance appears on both qubit and resonator drive channels)

        """
        phase_correction: dict[MoveMarker, float] = {}

        for resonator in builder.chip_topology.computational_resonators:
            r_drive_channel_name = builder.get_drive_channel(resonator)
            if r_drive_channel_name not in schedule:
                continue

            move_qubit: str | None = None
            rz_phase: float = 0  # accumulated phase correction
            dt_samples: int = 0  # duration of the MOVE sandwich in samples
            r_drive_channel = iter(schedule[r_drive_channel_name])
            for inst in r_drive_channel:
                if not isinstance(inst, (Wait, Block, VirtualRZ, MoveMarker)):
                    raise ValueError(f"{type(inst)} in virtual drive channel of {resonator}, should not happen.")

                if move_qubit:
                    # ongoing MOVE sandwich
                    if isinstance(inst, MoveMarker):
                        # move ends
                        # handle the RZ from the MOVE gate itself
                        rz_phase += handle_move_rz(next(r_drive_channel), resonator)
                        if apply_detuning_corrections:
                            if move_qubit != inst.qubit:
                                raise ValueError(
                                    f"""Resonator {resonator}: interleaved MOVE gates between {move_qubit} and
                                    {inst.qubit}"""
                                )
                            # dynamic phase change due to detuning
                            q_drive_channel_name = builder.get_drive_channel(move_qubit)
                            q_drive_channel = builder.channels[q_drive_channel_name]
                            duration = q_drive_channel.duration_to_seconds(dt_samples)
                            detuning_phase = 2 * np.pi * (inst.detuning * duration)
                        else:
                            detuning_phase = 0

                        phase_correction[inst] = rz_phase - detuning_phase
                        move_qubit = None
                    elif isinstance(inst, VirtualRZ):
                        rz_phase += inst.phase_increment
                        dt_samples += inst.duration
                    else:
                        # Wait or Block
                        dt_samples += inst.duration
                elif isinstance(inst, MoveMarker):
                    # move begins
                    move_qubit = inst.qubit
                    # handle the RZ from the MOVE gate itself
                    rz_phase = handle_move_rz(next(r_drive_channel), resonator)
                    dt_samples = 0
        return phase_correction

    phase_correction = find_phase_corrections()
    new_schedule: dict[str, list[Instruction]] = {}

    # apply the phase corrections on the drive channels of the moved qubits
    for qubit in builder.chip_topology.qubits:
        q_drive_channel_name = builder.get_drive_channel(qubit)
        if q_drive_channel_name not in schedule:
            continue

        # replace qubit drive channel contents, removing MoveMarker instructions and applying z rotations
        instructions: list[Instruction] = []
        move_resonator: str | None = None
        q_drive_channel = iter(schedule[q_drive_channel_name])
        for inst in q_drive_channel:
            if isinstance(inst, MoveMarker):
                if not move_resonator:
                    # move begins
                    move_resonator = inst.resonator
                else:
                    if apply_detuning_corrections and move_resonator != inst.resonator:
                        raise ValueError(
                            f"Qubit {qubit}: interleaved MOVE gates between {move_resonator} and {inst.resonator}"
                        )
                    # move ends
                    move_resonator = None
                    phase = phase_correction[inst]
                    # a VirtualRZ instruction should always follow the ending marker since the MOVE gate itself has one
                    vrz = next(q_drive_channel)
                    phase += handle_move_rz(vrz, qubit)
                    # replace the MOVE VirtualRZ with one applying the correct phase shift
                    # Normalize the phase increment to (-pi, pi] so that high number of full turns
                    # do not mess up the instruments (280+ full turns seem to cause some problems based
                    # on tests, probably due to IEEE 754 floating point overflows)
                    instructions.append(
                        VirtualRZ(
                            duration=vrz.duration,
                            phase_increment=normalize_angle(phase),
                        )
                    )
            else:
                instructions.append(inst)

        new_schedule[q_drive_channel_name] = instructions

    return Schedule(
        {ch: new_schedule[ch] if ch in new_schedule else list(instructions) for ch, instructions in schedule.items()}
    )


def validate_move_instructions(
    instructions: Iterable[CircuitOperation],
    builder: ScheduleBuilder,
    validate_prx: bool = True,
) -> Iterable[CircuitOperation]:
    """Circuit-level pass to prepare a circuit containing MOVE gates for compilation.

    Validates that circuit conforms to the MOVE gate constraints.

    Args:
        instructions: quantum circuit to validate
        builder: schedule builder, encapsulating information about the station
        validate_prx: whether to validate the circuit for PRX gates between MOVE sandwiches as well
    Returns:
        ``instructions``, unmodified
    Raises:
        ValueError: Circuit does not conform to MOVE constraints.

    """
    # Mapping from resonator to the qubit whose state was moved to it
    resonator_occupations: dict[str, str] = {}
    # Qubits whose states are currently moved to a resonator
    moved_qubits: set[str] = set()
    chip_topology = builder.chip_topology

    for inst in instructions:
        if inst.name == "move":
            qubit, resonator = inst.locus
            if not (chip_topology.is_qubit(qubit) and chip_topology.is_computational_resonator(resonator)):
                raise ValueError(f"Move operation locus must always be (qubit, resonator), got '{inst.locus}'")

            if (resonator_qubit := resonator_occupations.get(resonator)) is None:
                # Beginning MOVE: check that the qubit hasn't been moved to another resonator
                if qubit in moved_qubits:
                    raise ValueError(
                        f"Cannot apply MOVE{inst.locus} because "
                        + "the state of {qubit} is already moved to another resonator"
                    )
                resonator_occupations[resonator] = qubit
                moved_qubits.add(qubit)
            else:
                # Ending MOVE: check that the qubit matches to the qubit that was moved to the resonator
                if resonator_qubit != qubit:
                    raise ValueError(
                        f"Cannot apply MOVE{inst.locus} because "
                        + f"{resonator} already holds the state of '{resonator_qubit}'"
                    )
                del resonator_occupations[resonator]
                moved_qubits.remove(qubit)
        elif moved_qubits:
            # Validate that moved qubits are not used during MOVE operations
            if (inst.name != "barrier") and (validate_prx or inst.name != "prx"):
                # Barriers are allowed since they're just meta information and
                # not interacting directly with any real channels
                if overlap := set(inst.locus) & moved_qubits:
                    raise ValueError(
                        f"Operation {inst.name} to qubits '{overlap}' are forbidden while their states are moved to "
                        + "a resonator"
                    )

    # Finally validate that all moves have been ended before the circuit ends
    if resonator_occupations:
        raise ValueError(
            "The following resonators are still holding qubit states at "
            + f"the end of the circuit: {resonator_occupations}"
        )
    return instructions



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tools for building instruction schedules."""

from __future__ import annotations

from collections.abc import Iterable
import copy
from dataclasses import dataclass, field, replace
import itertools
import logging
from types import MethodType
from typing import Any

from iqm.models.playlist.channel_descriptions import (
    ChannelDescription,
    IQChannelConfig,
    ReadoutChannelConfig,
    RealChannelConfig,
)
import iqm.models.playlist.instructions as sc_instructions  # TODO make SC use iqm.pulse Instructions?
from iqm.models.playlist.segment import Segment as SC_Schedule  # iqm.models names are temporary
from iqm.models.playlist.waveforms import to_canonical

from exa.common.helpers.yaml_helper import load_yaml
from exa.common.qcm_data.chip_topology import ChipTopology, sort_components
from iqm.pulse.base_utils import _dicts_differ, merge_dicts
from iqm.pulse.gate_implementation import (
    CompositeCache,
    GateImplementation,
    Locus,
    OILCalibrationData,
    OpCalibrationDataTree,
)
from iqm.pulse.gates import _exposed_implementations
from iqm.pulse.gates.default_gates import _default_implementations, _quantum_ops_library
from iqm.pulse.playlist.channel import ChannelProperties, ProbeChannelProperties
from iqm.pulse.playlist.instructions import (
    AcquisitionMethod,
    ComplexIntegration,
    ConditionalInstruction,
    FluxPulse,
    Instruction,
    IQPulse,
    MultiplexedIQPulse,
    ReadoutTrigger,
    RealPulse,
    ThresholdStateDiscrimination,
    TimeTrace,
    VirtualRZ,
    Wait,
)
from iqm.pulse.playlist.playlist import Playlist
from iqm.pulse.playlist.schedule import Schedule, Segment
from iqm.pulse.quantum_ops import QuantumOp, QuantumOpTable, validate_locus_calibration, validate_op_calibration
from iqm.pulse.scheduler import (
    NONSOLID,
    Block,
    extend_hard_boundary,
    extend_hard_boundary_in_seconds,
    extend_schedule_new,
)
from iqm.pulse.timebox import SchedulingAlgorithm, SchedulingStrategy, TimeBox
from iqm.pulse.utils import (
    _process_implementations,
    _validate_locus_defaults,
    _validate_op_attributes,
)

logger = logging.getLogger(__name__)


@dataclass
class CircuitOperation:
    """Specific quantum operation applied on a specific part of the QPU, e.g. in a quantum circuit."""

    name: str
    """name of the quantum operation"""
    locus: tuple[str, ...]
    """names of the information-bearing QPU components (qubits, computational resonators...) the operation acts on"""
    args: dict[str, Any] = field(default_factory=dict)
    """arguments for the operation"""
    implementation: str | None = None
    """name of the implementation"""

    def validate(self, op_table: QuantumOpTable) -> None:
        """Validate the operation against a table of operation definitions.

        Args:
            op_table: table containing allowed quantum operations

        Raises:
            ValueError: operation is not valid

        """
        # find the op type
        op_type = op_table.get(self.name)
        if op_type is None:
            raise ValueError(f"Unknown quantum operation '{self.name}'.")

        # find the implementation
        impl_name = self.implementation
        if impl_name is not None and impl_name not in op_type.implementations:
            raise ValueError(f"Unknown implementation '{impl_name}' for quantum operation '{self.name}'.")

        if 0 < op_type.arity != len(self.locus):
            raise ValueError(
                f"The '{self.name}' operation acts on {op_type.arity} qubit(s), "
                f"but {len(self.locus)} were given: {self.locus}"
            )

        if not set(op_type.params).issubset(self.args):
            raise ValueError(
                f"The '{self.name}' operation requires the arguments {op_type.params}, "
                f"but {tuple(self.args)} were given."
            )


def load_config(path: str) -> tuple[QuantumOpTable, OpCalibrationDataTree]:
    """Load quantum operation definitions and calibration data from a YAML config file.

    Args:
        path: path to a YAML config file

    Returns:
        quantum operation definitions, calibration data tree

    """
    yaml = load_yaml(path)
    if (exp := yaml.get("experiment")) is None:
        raise ValueError('The config YAML is missing the "experiment" section.')

    ops = exp.get("gate_definitions", {})
    calibration = exp.get("gates", {})
    return build_quantum_ops(ops), calibration


def validate_quantum_circuit(
    operations: Iterable[CircuitOperation],
    op_table: QuantumOpTable,
    *,
    require_measurements: bool = False,
) -> None:
    """Validate a sequence of circuit operations constituting a quantum circuit.

    Args:
        operations: quantum circuit to be validated
        op_table: table containing allowed/calibrated quantum operations
        require_measurements: iff True the circuit must include at least one measurement operation

    Raises:
        ValueError: ``operations`` do not constitute a valid quantum circuit

    """
    measurement_keys: set[str] = set()
    measured_qubits: set[str] = set()

    for op in operations:
        op.validate(op_table)

        if key := op.args.get("key"):
            # this is a measurement operation
            if key in measurement_keys:
                raise ValueError(f"Measurement key '{key}' is not unique.")

            measurement_keys.add(key)
            for qubit in op.locus:
                measured_qubits.add(qubit)

    if require_measurements and not measured_qubits:
        raise ValueError("Circuit contains no measurements.")


def build_quantum_ops(ops: dict[str, Any]) -> QuantumOpTable:
    """Builds the table of known quantum operations.

    Hardcoded default native ops table is extended by the ones in `ops`.
    In case of name collisions, the content of `ops` takes priority over the defaults.

    Args:
        ops: Contents of the ``gate_definitions`` section defining
        the quantum operations in the configuration YAML file.
        Modified by the function.

    Returns:
        Mapping from quantum operation name to its definition

    """
    op_table = copy.deepcopy(_quantum_ops_library)
    for op_name, definition in ops.items():
        implementations_def = definition.pop("implementations", {})
        implementations = _process_implementations(
            op_name, implementations_def, _exposed_implementations, _default_implementations
        )

        if "defaults_for_locus" in definition:
            _validate_locus_defaults(op_name, definition, implementations)

        if op_name in op_table:
            _validate_op_attributes(op_name, definition, op_table)
            op_table[op_name] = replace(op_table[op_name], implementations=implementations, **definition)
        else:
            op_table[op_name] = QuantumOp(name=op_name, implementations=implementations, **definition)

    return op_table


class ScheduleBuilder:
    """Builds instruction schedules out of quantum circuits or individual quantum operations.

    Encapsulates known quantum ops, the calibration information for them, QPU components and their
    topology, and controller properties.

    Args:
        op_table: definitions of known quantum ops
        calibration: calibration data tree for the quantum ops
        chip_topology: Chip topology derived from the CHAD.
        channels: mapping of controller names to the configurations of their channels
        component_channels: Mapping from QPU component name to a mapping of ``('drive', 'flux', 'readout')``
            to the name of the control channel responsible for that function of the component.

    """

    def __init__(
        self,
        op_table: QuantumOpTable,
        calibration: OpCalibrationDataTree,
        chip_topology: ChipTopology,
        channels: dict[str, ChannelProperties],
        component_channels: dict[str, dict[str, str]],
    ):
        # TODO: channels and component_channels should be given in a single Protocol here.
        self.op_table = op_table
        self.calibration = calibration
        self.chip_topology = chip_topology
        self.channels = channels
        self.component_channels = component_channels
        self._cache: dict[str, dict[str, dict[Locus, GateImplementation]]] = {}
        """Cached GateImplementations. The tree has the same structure as OpCalibrationDataTree."""
        self.composite_cache = CompositeCache()
        """Cache for the CompositeGate TimeBoxes. Flushed whenever ANY calibration data is injected into the builder.
        """
        self._channel_to_component: dict[str, str] = {
            c: q
            for q, chans in self.component_channels.items()
            for c in chans.values()
            if self.channels[c].blocks_component
        }
        """``self.component_channels`` mapping inverted cached for scheduling algorithm performance. This mapping is
        used in the scheduling to determine the components to block based on their associated channels. Only
        blocking channels are included in this mapping, non-blocking channels (e.g. certain virtual channels) do not
        block their components, just themselves."""
        self._channel_types = {
            "other": [c for c, prop in self.channels.items() if not isinstance(prop, ProbeChannelProperties)],
            "probe": [c for c, prop in self.channels.items() if isinstance(prop, ProbeChannelProperties)],
        }
        """Cache the probe and non-probe channel names for the scheduling algorithm performance"""
        self._require_scheduling_in_seconds: bool = len(
            {c.sample_rate for c in self.channels.values() if not c.is_virtual}
        ) > 1 or {
            self.channels[c].instruction_duration_granularity
            for c in self._channel_types["other"]
            if not self.channels[c].is_virtual
        } != {
            self.channels[c].instruction_duration_granularity
            for c in self._channel_types["probe"]
            if not self.channels[c].is_virtual
        }
        """Whether to require scheduling of probe instructions in seconds instead of in samples. This can happen for two
        reasons: 1) the probe channel has a different sampling rate to the other channels (e.g. with UHFQA) or
        2) the probe channels have a different instruction granularity to some of the other channels (e.g. with
        mixed stations that have the RO device from a different vendor than some of the AWGs)."""
        self._channel_templates: dict[str, ChannelProperties | None] = {
            "other": next((c for k, c in self.channels.items() if k in self._channel_types["other"]), None),
            "probe": next((c for k, c in self.channels.items() if k in self._channel_types["probe"]), None),
        }
        """Cache representative channel properties for a probe and a non-probe channel for the scheduling algorithm
        performance."""

        for op_name in op_table:
            self._set_gate_implementation_shortcut(op_name)

        self._logger = logging.getLogger(__name__)

    def __getitem__(self, item: str) -> GateImplementation:
        """If ``item`` is an operation name in ``self.op_table``, returns the shortcut method ``self.<op_name>``"""
        if item in self.op_table:
            return getattr(self, item)
        raise ValueError(f"No operation found with the name {item} in ``self.op_table``.")

    def inject_calibration(self, partial_calibration: OpCalibrationDataTree) -> None:
        """Inject new calibration data, changing ``self.calibration`` after the ScheduleBuilder initialisation.

        Invalidates the gate_implementation cache for the affected operations/implementations/loci. Also invalidates
        the cache for any factorizable gate implementation, if any of its locus components was affected.

        Args:
            partial_calibration: data to be injected. Must have the same structure as ``self.calibration`` but does not
                have to contain all operations/implementations/loci/values. Only the parts of the data that are
                found will be merged into ``self.calibration`` (including any ``None`` values). ``self._cache`` will
                be invalidated for the found operations/implementations/loci and only if the new calibration data
                actually differs from the previous.

        """
        self.composite_cache.flush()
        for op, op_data in partial_calibration.items():
            for impl, impl_data in op_data.items():
                for locus, locus_data in impl_data.items():
                    prev_calibration = self.calibration[op][impl][locus]
                    new_calibration = merge_dicts(prev_calibration, locus_data)
                    self.calibration[op][impl][locus] = new_calibration
                    if (
                        op in self._cache
                        and impl in self._cache[op]
                        and locus in self._cache[op][impl]
                        and _dicts_differ(prev_calibration, new_calibration)
                    ):
                        del self._cache[op][impl][locus]
                        if self.op_table[op].factorizable:
                            factorizable_cache = self._cache[op][impl].copy()
                            set_locus = set(locus)
                            for cache_locus in factorizable_cache:
                                if set_locus.intersection(set(cache_locus)):
                                    del self._cache[op][impl][cache_locus]

    def validate_calibration(self) -> None:
        """Check that the calibration data matches the known quantum operations.

        Raises:
            ValueError: there is something wrong with the calibration data

        """
        validate_op_calibration(self.calibration, self.op_table)

    def get_drive_channel(self, component: str) -> str:
        """Drive channel for the given QPU component.

        Args:
            component: name of a QPU component
        Returns:
            Name of the drive channel for ``component``, if it exists.

        Raises:
            KeyError: if component does not exist or does not have a drive channel

        """
        return self._get_channel_for_component(component, "drive")

    def get_flux_channel(self, component: str) -> str:
        """Flux channel for the given QPU component.

        See :meth:`.get_drive_channel`.
        """
        return self._get_channel_for_component(component, "flux")

    def get_probe_channel(self, component: str) -> str:
        """Probe line channel for the probe line ``component`` belongs to.

        See :meth:`.get_drive_channel`.
        """
        probe_line = self.chip_topology.component_to_probe_line.get(component, None)
        if probe_line is None:
            raise KeyError(f'probe line not found for component "{component}"')
        return self._get_channel_for_component(probe_line, "readout")

    def get_virtual_feedback_channels(self, component: str) -> list[str]:
        """All virtual feedback signal channels for the given QPU component.

        A virtual feedback channel between a source and a destination exists if the station configuration allows it.
        `component` can be either the source or the destination of the signal.

        Args:
            component: name of a QPU component
        Returns:
            Names of the virtual channels.

        """
        return [
            channel_name
            for operation, channel_name in self.component_channels[component].items()
            if operation.startswith("feedback")
        ]

    def get_virtual_feedback_channel_for(self, awg_name: str, feedback_qubit: str) -> str:
        """Get virtual feedback channel for feedback to a given AWG from a given probe line.

        Args:
            awg_name: name of the awg node that receives the feedback bit.
            feedback_qubit: which qubit's measurement resulted in the feedback bit

        Returns:
            The virtual feedback channel name.

        Raises:
            ValueError: if the given AWG does not support fast feedback from the given probe line.

        """
        probe_line = self.chip_topology.component_to_probe_line.get(feedback_qubit)
        all_virtual_channels = self.get_virtual_feedback_channels(probe_line)
        channel = next((c for c in all_virtual_channels if awg_name in c), None)
        if not channel:
            raise ValueError(f"AWG node {awg_name} does not support fast feedback from {probe_line}")
        return channel

    def _get_channel_for_component(self, component: str, operation: str) -> str:
        """Control channel name for the given QPU component and operation.

        Returns:
            name of the channel
        Raises:
            KeyError: if component does not exist or does not have the operation

        """
        if (
            self.component_channels is None
            or component not in self.component_channels
            or operation not in self.component_channels[component]
        ):
            raise KeyError(f'channel not found for component "{component}" and operation "{operation}"')
        return self.component_channels[component][operation]

    def has_calibration(
        self,
        op_name: str,
        impl_name: str,
        locus: Locus,
    ) -> bool:
        """Is there calibration data for the given quantum operation, implementation and locus?

        Args:
            op_name: name of the quantum operation
            impl_name: name of the implementation
            locus: locus of the operation

        Returns:
            True iff requested calibration data was found

        """
        if (op_cal := self.calibration.get(op_name)) is None:
            return False
        if (impl_cal := op_cal.get(impl_name)) is None:
            return False
        if locus not in impl_cal:
            # if a locus node is missing entirely, is not replaced with the defaults
            return False
        return True

    def get_calibration(
        self,
        op_name: str,
        impl_name: str,
        locus: Locus,
    ) -> OILCalibrationData:
        """Calibration data for the given quantum operation, implementation and locus.

        Args:
            op_name: name of the quantum operation
            impl_name: name of the implementation
            locus: locus of the operation

        Returns:
            requested calibration data

        Raises:
            ValueError: requested calibration data was not found

        """
        if (op_cal := self.calibration.get(op_name)) is None:
            raise ValueError(f"No calibration data for op '{op_name}'.")
        if (impl_cal := op_cal.get(impl_name)) is None:
            raise ValueError(f"No calibration data for '{op_name}.{impl_name}'.")
        if (cal_data := impl_cal.get(locus)) is None:
            # if a locus node is missing entirely, is not replaced with the defaults
            raise ValueError(f"No calibration data for '{op_name}.{impl_name}' at {locus}.")

        default_cal_data = impl_cal.get((), {})  # empty locus contains the default cal data for all the loci
        return merge_dicts(default_cal_data, cal_data, merge_nones=False)

    def get_control_channels(
        self,
        locus: Iterable[str],
    ) -> tuple[str, ...]:
        """Control channels that directly affect quantum operations at the given locus.

        Includes the probe, drive and flux channels of the locus QPU components.
        Does not include e.g. any neighboring coupler channels, these will have to be added
        separately in the TimeBox resolution phase.

        Will only return channels that are known to exist, i.e. are found in :attr:`ScheduleBuilder.channels`.

        Args:
            locus: locus on which the operation acts

        Returns:
            names of the control channels that directly affect the operation

        """
        controllers: list[str] = []
        for q in locus:
            if self.component_channels is not None and q in self.component_channels:
                controllers.extend(self.component_channels[q].values())
        return tuple(c for c in controllers if c in self.channels and not self.channels[c].is_virtual)

    def wait(self, locus: Iterable[str], duration: float, *, rounding: bool = False) -> TimeBox:
        """Utility method for applying Block instructions on every channel of the given locus.

        The Block instructions guarantee the locus components to idle for the given duration,
        and cannot e.g. be replaced with e.g. dynamical decoupling sequences.
        They are treated the same as any other TimeBox contents:

        1. Blocks on different channels remain aligned in time during scheduling.
        2. The actual waiting time on a particular channel may thus be >= ``duration``,
           if the other channels have less non-blocking space on either side.

        .. note::
           TODO For now, this method can round ``duration`` to the nearest value allowed by each
           channel if requested. This is for the benefit of EXA sweeping over waiting durations.
           In the future, EXA sweep generation should be responsible for doing the rounding.

        Args:
            locus: locus components that should experience the wait
            duration: how long to wait (in seconds)
            rounding: Iff True, for each channel separately, ``duration`` will be rounded to the
                nearest value allowed by the granularity of that channel. The Waits will start
                simultaneously.

        Returns:
            box containing :class:`.Block` instructions on every control channel of ``locus``

        """
        channels = self.get_control_channels(locus)
        # Zero-duration wait is equivalent to doing nothing on the channels, however
        # it will act as a barrier during scheduling.
        if duration == 0:
            segments = {ch: [Block(0)] for ch in channels}
        else:
            segments = {}
            for channel in channels:
                if rounding:
                    rounded = self.channels[channel].round_duration_to_granularity(duration)
                    int_duration = self.channels[channel].duration_to_int_samples(rounded) if duration > 0 else 0
                else:
                    int_duration = self.channels[channel].duration_to_int_samples(duration)
                segments[channel] = [Block(int_duration)]

        return TimeBox.atomic(
            Schedule(segments),
            locus_components=locus,
            label="Wait",
        )

    def get_implementation(
        self,
        op_name: str,
        locus: Iterable[str],
        impl_name: str | None = None,
        *,
        use_priority_order: bool = False,
        strict_locus: bool = False,
        priority_calibration: OILCalibrationData | None = None,
    ) -> GateImplementation:
        """Provide an implementation for a quantum operation at a given locus.

        Args:
            op_name: name of the quantum operation
            locus: locus of the operation
            impl_name: name of the implementation (``None`` means the implementation is chosen automatically
                using the logic described below)
            strict_locus: iff False, for non-symmetric implementations of symmetric ops the locus order may
                be changed if no calibration data is available for the requested locus order
            use_priority_order: Only has an effect if ``impl_name`` is ``None``. Iff ``False``,
                :meth:`QuantumOp.get_default_implementation_for_locus` is used. Otherwise, the first implementation in
                the priority order that has calibration data for ``locus`` is chosen. The priority order is as follows:
                1. The locus-specific priority defined in ``QuantumOp.defaults_for_locus[locus]`` if any.
                2. The global priority order defined in :attr:`QuantumOp.implementations`.
            priority_calibration: Calibration data from which to load the calibration instead of the common calibration
                data in :attr:`calibration`. If no calibration is found for the given implementation or
                ``priority_calibration`` is ``None``, the common calibration is used. Any non-empty
                values found in ``priority_calibration`` will be merged to the common calibration. Note:
                using ``priority_calibration`` will prevent saving/loading via the cache.

        Returns:
            requested implementation

        Raises:
            ValueError: requested implementation could not be provided

        """
        if (op := self.op_table.get(op_name)) is None:
            raise ValueError(f"Unknown quantum operation '{op_name}'.")

        if impl_name is None and not use_priority_order:
            impl_name = op.get_default_implementation_for_locus(locus)
        return self._get_implementation(
            op, impl_name, tuple(locus), strict_locus, priority_calibration=priority_calibration
        )

    def _find_implementation_and_locus(
        self,
        op: QuantumOp,
        impl_name: str | None,
        locus: Locus,
        *,
        strict_locus: bool = False,
    ) -> tuple[str, Locus]:
        """Find an implementation and locus for the given quantum operation instance compatible
        with the calibration data.

        Args:
            op: quantum operation
            impl_name: Name of the implementation. ``None`` means use the highest-priority implementation for
                which we have calibration data.
            locus: locus of the operation
            strict_locus: iff False, for non-symmetric implementations of symmetric ops the locus order may
                be changed to an equivalent one if no calibration data is available for the requested locus order

        Returns:
            chosen implementation name, locus

        Raises:
            ValueError: requested implementation could not be found
            ValueError: requested implementation had no calibration data for this locus
            ValueError: no specific implementation was requested, but no known implementation had
                calibration data for this locus

        """

        def find_locus(
            op: QuantumOp,
            impl_name: str,
            impl_class: type[GateImplementation],
            given_locus: Locus,
            strict_locus: bool = False,
        ) -> Locus | None:
            """Tries to find a valid locus for the given operation and implementation, that has
            calibration data available. If none can be found, returns None.
            """
            if not impl_class.needs_calibration():
                return given_locus
            if op.factorizable and len(given_locus) > 1:
                return given_locus
            # find out which loci we need to check for cal data
            if op.symmetric:
                if impl_class.symmetric:
                    # Cal data for symmetric implementations uses always a sorted locus order.
                    loci = [tuple(sort_components(given_locus))]
                elif strict_locus:
                    # If the operation is symmetric but implementation is not (e.g. fast flux CZ)
                    # the locus order can be meaningful.
                    # Users must be able to request implementations for any order of the locus, which may have
                    # independent cal data.
                    # User requested this implementation, we assume they also want this particular locus.
                    loci = [given_locus]
                else:
                    # User did not request a specific implementation, so we assume they are fine
                    # with any implementation and locus order. We pick the first locus order that has cal data.
                    loci = list(itertools.permutations(given_locus))
            else:
                # For non-symmetric ops the locus is always strict.
                loci = [given_locus]

            for new_locus in loci:
                # use the first locus under which we can find cal data
                if self.has_calibration(op.name, impl_name, new_locus):
                    return new_locus
            return None

        def merge_impl_priorities_for_locus(op: QuantumOp, locus: Locus) -> dict[str, type[GateImplementation]]:
            """Merge locus-specific and global implementation priorities for a given locus. The implementation
            defined for the locus will take precedence over the global ones.
            """
            if not op.defaults_for_locus:
                return op.implementations
            locus_priority = op.get_default_implementation_for_locus(locus)
            priorities: dict[str, type[GateImplementation]] = {locus_priority: op.implementations[locus_priority]}
            for global_priority, priority_class in op.implementations.items():
                if global_priority not in priorities:
                    priorities[global_priority] = priority_class
            return priorities

        if impl_name is None:
            # no specific implementation requested, choose the highest-priority implementation
            # for which there is cal data and take into account locus-specific implementation order
            priorities = merge_impl_priorities_for_locus(op, locus)
            for new_impl_name, impl in priorities.items():
                if (
                    not impl.special_implementation
                    and (new_locus := find_locus(op, new_impl_name, impl, locus, strict_locus)) is not None
                ):
                    # valid implementation and locus found
                    return new_impl_name, new_locus
            # no valid implementation found
            raise ValueError(f"No calibration data for '{op.name}' at {locus}.")

        # use the requested implementation if it exists
        if (requested_impl := op.implementations.get(impl_name)) is None:
            raise ValueError(f"Unknown quantum operation implementation '{op.name}.{impl_name}'.")
        if (cal_locus := find_locus(op, impl_name, requested_impl, locus, strict_locus)) is None:
            raise ValueError(f"No calibration data for '{op.name}.{impl_name}' at {locus}.")
        return impl_name, cal_locus

    def _get_implementation(
        self,
        op: QuantumOp,
        impl_name: str | None,
        locus: Locus,
        strict_locus: bool = False,
        priority_calibration: dict[str, Any] | None = None,
    ) -> GateImplementation:
        """Build a factory class for the given quantum operation, implementation and locus.

        The GateImplementations are built when they are first requested, and cached for later use.

        Args:
            op: quantum operation
            impl_name: Name of the implementation. ``None`` means use the highest-priority implementation for
                which we have calibration data.
            locus: locus of the operation
            strict_locus: iff False, for non-symmetric implementations of symmetric ops the locus order may
                be changed if no calibration data is available for the requested locus order
            priority_calibration: Calibration data from which to load the calibration instead of the common
                calibration data. Priority calibration should be either a dict of the type `OILCalibrationData`,
                i.e. containing the operation name, implementation name, and locus, or just a dict containing
                the calibration data for the locus implied by the args `op`, `impl_name` and `locus`.

        Returns:
            requested implementation

        Raises:
            ValueError: requested implementation could not be provided or had no calibration data for this locus

        """
        new_impl_name, new_locus = self._find_implementation_and_locus(op, impl_name, locus, strict_locus=strict_locus)

        # use a cached factory if it exists and no priority calibration is used:
        if priority_calibration is None:
            op_cache = self._cache.setdefault(op.name, {})
            impl_cache = op_cache.setdefault(new_impl_name, {})
            if factory := impl_cache.get(new_locus):
                return factory

        impl_class = op.implementations[new_impl_name]
        if impl_class.needs_calibration() and (not op.factorizable or (op.factorizable and len(new_locus) == 1)):
            # find the calibration data
            cal_data = self.get_calibration(op.name, new_impl_name, new_locus)
            if priority_calibration:
                if op.name in priority_calibration:
                    # first check if full OILCalibrationData structure was given
                    op_data = priority_calibration.get(op.name, {})
                    prio_data = op_data.get(new_impl_name, {}).get(new_locus, {})
                else:
                    # if not assume just the data for this locus was given
                    prio_data = priority_calibration
                cal_data = merge_dicts(cal_data, prio_data, merge_nones=False) if prio_data else cal_data
            validate_locus_calibration(cal_data, impl_class, op, new_impl_name, new_locus)
        elif op.factorizable and len(new_locus) > 1 and priority_calibration:
            # only calibration data a factorizable gate needs for a multi-qubit locus is possible prio calibration
            # we propagate it to the gate implementation which should take care of handling it correctly
            cal_data = priority_calibration
        else:
            cal_data = {}

        # construct the factory
        factory = impl_class(
            parent=op,
            name=new_impl_name,
            locus=new_locus,
            calibration_data=cal_data,
            builder=self,
        )
        # cache it if no priority_calibration was used
        if priority_calibration is None:
            impl_cache[new_locus] = factory
        return factory

    def get_implementation_class(self, op_name: str, impl_name: str | None = None) -> type[GateImplementation]:
        """Implementation class for the given operation.

        Args:
            op_name: name of the quantum operation
            impl_name: name of the implementation (``None`` means use the default implementation)

        Returns:
            requested implementation class

        """
        op = self.op_table[op_name]
        if impl_name is None:
            if not op.implementations:
                raise ValueError(f"No implementations found for operation {op_name}.")
            return next(iter(op.implementations.values()))
        return op.implementations[impl_name]

    def validate_quantum_circuit(
        self,
        operations: Iterable[CircuitOperation],
        *,
        require_measurements: bool = False,
    ) -> None:
        """Validate a sequence of circuit operations constituting a quantum circuit.

        Args:
            operations: quantum circuit to be validated
            require_measurements: iff True the circuit must include at least one measurement operation

        Raises:
            ValueError: ``operations`` do not constitute a valid quantum circuit

        """
        validate_quantum_circuit(operations, self.op_table, require_measurements=require_measurements)

    def circuit_to_timebox(
        self,
        circuit: Iterable[CircuitOperation],
        *,
        name: str = "",
        scheduling_algorithm: SchedulingAlgorithm = SchedulingAlgorithm.HARD_BOUNDARY,
        locus_mapping: dict[str, str] | None = None,
    ) -> TimeBox:
        """Convert a quantum circuit to a TimeBox.

        Args:
            circuit: quantum circuit
            name: name of the circuit
            scheduling_algorithm: scheduling algorithm to be used in resolving the TimeBoxes.
            locus_mapping: optional mapping of placeholder component names to the physical component names used
                while resolving the circuit into a TimeBox.

        Returns:
            unresolved TimeBox that implements ``circuit``

        Raises:
            ValueError: failed to convert ``circuit`` to a TimeBox

        """
        # TODO for now do not force validation here
        # self.validate_quantum_circuit(circuit, require_measurements=True)

        boxes = []
        locus_mapping = locus_mapping or {}
        for op in circuit:
            self._logger.debug("Adding %s", op)

            # we have already validated the operations, so we know they can be found in self.op_table
            op_type = self.op_table[op.name]
            # append the operation to the pulse schedule
            # we operate in non-strict mode, i.e. for symmetric gates the caller does not
            # need to know the locus order, any will do
            mapped_locus = tuple(locus_mapping.get(qubit, qubit) for qubit in op.locus) if locus_mapping else op.locus
            factory = self._get_implementation(op_type, op.implementation, mapped_locus)
            boxes.append(factory(**op.args))
        return TimeBox.composite(boxes, label=name, scheduling_algorithm=scheduling_algorithm)

    def timeboxes_to_front_padded_playlist(self, boxes: Iterable[TimeBox], *, neighborhood: int = 0) -> Playlist:
        """Temporary helper function, for converting a sequence of TimeBoxes to a Playlist.

        Each individual TimeBox in ``boxes`` is resolved into a Schedule, and then
        each schedules is front-padded with :class:`.Wait` instructions on each channel
        such that the resulting Schedules have equal durations. This is required since
        for now in Station Control the delay before the final measurement is the same for
        all the Schedules in a Playlist, and we do not wish to lose coherence waiting for
        the measurement after each Schedule is done.

        TODO Once Station Control can handle measurements better, this method should be removed,
        and :meth:`timeboxes_to_playlist` be used instead.

        Args:
            boxes: TimeBoxes to include in the playlist
            neighborhood: During scheduling, block neighboring channels of the used components this far. By default,
                blocks only the defined locus components and any other components which have occupied channels.

        Returns:
            playlist that implements ``boxes``

        """
        schedules = [self.resolve_timebox(box, neighborhood=neighborhood).cleanup() for box in boxes]
        max_duration_in_seconds = max(
            (
                self.channels[ch].duration_to_seconds(seg.duration)
                for schedule in schedules
                for ch, seg in schedule.items()
            ),
            default=0.0,
        )
        T = max((schedule.duration for schedule in schedules), default=0)
        # We can assume every segment has a probe pulse (readout), but let's check the first one just in case
        if self._require_scheduling_in_seconds and schedules[0].has_content_in(self._channel_types["probe"]):
            schedules = [
                schedule.front_pad_in_seconds(max_duration_in_seconds, self.channels) for schedule in schedules
            ]
        else:
            schedules = [schedule.front_pad(T) for schedule in schedules]
        return self.build_playlist(schedules)

    def timeboxes_to_playlist(
        self,
        boxes: Iterable[TimeBox],
        *,
        neighborhood: int = 1,
    ) -> Playlist:
        """Convert a sequence of TimeBoxes to a Playlist.

        Resolves the boxes, converts them to Schedules, removes unnecessary channels, and then packs
        the Schedules into a Playlist. Assumes all the TimeBoxes refer to the same QPU and its control channels.

        Args:
            boxes: TimeBoxes to include in the playlist
            neighborhood: During scheduling, block neighboring channels of the used components this far.
                The default value ensures that quantum operations work as intended, assuming the station
                is properly calibrated. Higher values may help defend against crosstalk, at the expense
                of a longer instruction schedule and thus more decoherence.

        Returns:
            playlist that implements ``boxes``

        """
        return self.build_playlist(self.timebox_to_schedule(box, neighborhood=neighborhood) for box in boxes)

    def timebox_to_schedule(
        self,
        box: TimeBox,
        *,
        neighborhood: int = 1,
    ) -> Schedule:
        """Convert a TimeBox to a finished instruction schedule, ready for execution.

        Resolves the box, then converts the durations of the instructions in the schedule to samples
        at the channel sample_rate.

        Args:
            box: TimeBox to resolve
            neighborhood: During scheduling, block neighboring channels of the used components this far.
                The default value ensures that quantum operations work as intended, assuming the station
                is properly calibrated. Higher values may help defend against crosstalk, at the expense
                of a longer instruction schedule and thus more decoherence.

        Returns:
            finished schedule that implements ``box``

        """
        schedule = self.resolve_timebox(box, neighborhood=neighborhood)
        return self._finish_schedule(schedule)

    def _finish_schedule(self, schedule: Schedule) -> Schedule:
        """Finishes the instruction schedule.

        * removes channels that only have Waits in them
        * fuses long-distance Rz corrections to the correct drive pulses

        Args:
            schedule: schedule to finish

        Returns:
            finished copy of ``schedule``

        """
        has_long_distance_rzs = False
        for channel in schedule.channels():
            if "flux" in channel:
                for inst in schedule[channel]:
                    if isinstance(inst, FluxPulse) and inst.rzs:
                        has_long_distance_rzs = True
                        break
            if has_long_distance_rzs:
                return self._fuse_long_distance_virtual_rzs(schedule)
        return schedule.cleanup()

    def _fuse_long_distance_virtual_rzs(self, schedule: Schedule) -> Schedule:
        """Fuse long-distance (i.e. out-of-gate-locus) VirtualRZ corrections with the next drive pulse
        happening after the FluxPulse they are correcting.

        """

        def fuse(inst: Instruction, idx: int, rz_angle: float, segment: Segment) -> None:
            """Fuse a VirtualRZ angle to an IQPulse."""
            if isinstance(inst, (VirtualRZ, IQPulse)):
                segment._instructions[idx] = replace(inst, phase_increment=inst.phase_increment + rz_angle)
            else:
                raise ValueError(f"Unknown drive channel instruction {inst}")

        schedule_copy = schedule.copy()
        for channel in [ch for ch in schedule.channels() if "flux" in ch]:
            sample_counter = 0
            for inst in schedule[channel]:
                if isinstance(inst, FluxPulse):
                    rzs = {ch: angle for ch, angle in inst.rzs if ch in schedule}
                    for drive_channel, rz_angle in rzs.items():
                        drive_sample_counter = 0
                        for drive_inst_idx, drive_inst in enumerate(schedule[drive_channel]):
                            if (
                                not isinstance(drive_inst, Wait)
                                and not isinstance(drive_inst, NONSOLID)
                                and drive_sample_counter >= sample_counter
                            ):
                                fuse(
                                    schedule_copy[drive_channel][drive_inst_idx],
                                    drive_inst_idx,
                                    -rz_angle,  # invert the angle as per the agreed convention
                                    schedule_copy[drive_channel],
                                )
                                break
                            drive_sample_counter += drive_inst.duration
                sample_counter += inst.duration
        return schedule_copy.cleanup()

    def resolve_timebox(
        self, box: TimeBox, *, neighborhood: int, compute_neighborhood_hard_boundary: bool = False
    ) -> Schedule:
        """Resolve a TimeBox.

        Resolves recursively each of the children of the box, and then concatenates the resulting
        Schedules into a new one using a specific scheduling strategy and algorithm.

        The supported algorithms are ``HARD_BOUNDARY``, which treats each composite TimeBox as a solid rectangle (the
        longest channel within defines the duration) and ``TETRIS``, which packs the schedule as tightly as possible
        (solid instructions still cannot overlap) regardless of the TimeBox boundaries.

        Modifies ``box`` so that it becomes atomic, if it isn't already.

        Args:
            box: TimeBox to resolve
            neighborhood: During scheduling, block control channels of neighboring QPU components this far
                from the locus. Values higher than 0 may help defend against crosstalk, at the expense
                of a longer instruction schedule and thus more decoherence.
            compute_neighborhood_hard_boundary: Whether to precompute the neighborhood components while resolving
                a composite ``TimeBox`` in the ``HARD_BOUNDARY`` algorithm. Typically one does not want to do this on
                the top layer composite ``TimeBox``, since it would be unused. The algorithm sets this ``True`` on
                lower layers, where it improves the performance as the neighborhood components are needed in scheduling.

        Returns:
            instruction schedule that implements ``box``

        """
        if box.scheduling_algorithm == SchedulingAlgorithm.HARD_BOUNDARY:
            return self._resolve_timebox_hard_boundary(
                box, neighborhood=neighborhood, compute_neighborhood=compute_neighborhood_hard_boundary
            )
        return self._resolve_timebox_tetris(box, neighborhood=neighborhood)

    def _resolve_timebox_hard_boundary(
        self, box: TimeBox, neighborhood: int, compute_neighborhood: bool = False
    ) -> Schedule:
        """Resolves a TimeBox using the ``HARD_BOUNDARY`` algorithm, which treats each composite TimeBox as a solid
        rectangle (the longest channel within defines the duration).
        """
        if box.atom is not None:
            # already resolved
            return box.atom

        self._logger.debug("\nResolving '%s':", box.label)
        schedule = Schedule()
        if compute_neighborhood:
            for dist in range(neighborhood + 1):
                box.neighborhood_components.setdefault(dist, set())

        # If variable sampling rates are detected, we need to start scheduling in seconds.
        # Until then, the scheduling uses samples, and all time durations are integers.
        component_durations: dict[str, int] = {}
        component_durations_seconds: dict[str, float] = {}
        scheduling_in_seconds = False

        child_order = reversed(box.children) if box.scheduling == SchedulingStrategy.ALAP else box.children
        for child in child_order:
            child_schedule = self.resolve_timebox(
                child, neighborhood=neighborhood, compute_neighborhood_hard_boundary=True
            )
            if (
                self._require_scheduling_in_seconds
                and not scheduling_in_seconds
                and child_schedule.has_content_in(self._channel_types["probe"])
            ):
                # convert the locus durations to seconds (no probe channels have been used so far)
                scheduling_in_seconds = True
                if self._channel_templates["other"] is not None:
                    component_durations_seconds = {
                        component: self._channel_templates["other"].duration_to_seconds(duration)
                        for component, duration in component_durations.items()
                    }

            child_components = self._get_neighborhood_hard_boundary(child, 0)
            neighborhood_components = self._get_neighborhood_hard_boundary(child, neighborhood)
            if scheduling_in_seconds:
                extend_hard_boundary_in_seconds(
                    schedule,
                    child_schedule,
                    child_components,
                    neighborhood_components,
                    component_durations_seconds,
                    box.scheduling == SchedulingStrategy.ALAP,
                    self.channels,
                )
            else:
                extend_hard_boundary(
                    schedule,
                    child_schedule,
                    child_components,
                    neighborhood_components,
                    component_durations,
                    box.scheduling == SchedulingStrategy.ALAP,
                )
            if compute_neighborhood:
                box.neighborhood_components[0].update(child_components)
                if neighborhood > 0:
                    box.neighborhood_components[neighborhood].update(neighborhood_components)
            if self._logger.isEnabledFor(logging.DEBUG):
                self._logger.debug("Updated schedule:\n%s", schedule.pprint())

        if scheduling_in_seconds:
            schedule.pad_to_hard_box_in_seconds(self.channels)
        else:
            schedule.pad_to_hard_box()

        if box.scheduling == SchedulingStrategy.ALAP:
            schedule = schedule.reverse_hard_box()

        self._logger.debug("'%s' resolved.", box.label)
        box.atom = schedule
        return schedule

    def _get_neighborhood_hard_boundary(self, box: TimeBox, neighborhood: int) -> set[str]:
        """Computes and caches the blocking neighborhoods for HARD_BOUNDARY algorithm.

        Args:
            box: Atomic TimeBox whose neighborhood to compute.
            neighborhood: Return QPU components this far from the locus.

        Returns:
            QPU components (plus maybe channels?) belonging the the given neighborhood of ``box``.

        """
        # check the cache
        if neighborhood in box.neighborhood_components:
            return box.neighborhood_components[neighborhood]

        # Box may have other channels beside the ones that belong to its locus (e.g. coupler flux for CZ gates).
        # We need to find the associated components in order to block them.
        # Cache the found components in case they are not yet cached.
        # In case the channel is not blocking (i.e. not found in ``channel_to_component``), we just
        # block the channel itself
        components = box.neighborhood_components.get(0)
        if components is None:
            components = box.locus_components.copy()
            for channel in box.atom:
                components.add(self._channel_to_component.get(channel, channel))
            box.neighborhood_components[0] = components

        # iteratively find and cache the higher neighborhoods
        for nh_index in range(1, neighborhood + 1):
            prev = components
            components = box.neighborhood_components.get(nh_index)
            if components is None:
                if nh_index % 2 == 1:
                    components = prev | self.chip_topology.get_neighbor_couplers(prev)
                else:
                    components = prev | self.chip_topology.get_neighbor_locus_components(prev)
                box.neighborhood_components[nh_index] = components

        return components

    def _resolve_timebox_tetris(
        self,
        box: TimeBox,
        *,
        neighborhood: int,
    ) -> Schedule:
        """Resolves a TimeBox using the ``TETRIS`` algorithm, which packs the schedule as tightly as possible
        (solid instructions still cannot overlap) regardless of the TimeBox boundaries.
        """
        raise NotImplementedError(
            "SchedulingAlgorithm.TETRIS is not supported in this iqm-pulse version. "
            "Updating it to work with integer-duration scheduling will be done in COMP-1281."
        )

        if box.atom is not None:
            # already resolved
            return box.atom

        self._logger.debug("\nResolving '%s':", box.label)
        schedule = Schedule()

        # ALAP scheduling strategy works by reversing both the contents and order of the children before extending,
        # and then reversing the result schedule in the end.
        child_order = reversed(box.children) if box.scheduling == SchedulingStrategy.ALAP else box.children
        for child in child_order:
            child_schedule = self.resolve_timebox(child, neighborhood=neighborhood)
            self._block_neighborhood_tetris(child_schedule, child.locus_components, neighborhood)

            if self._logger.isEnabledFor(logging.DEBUG):  # evaluate pprint only if needed
                self._logger.debug("Adding %s\n%s", child.label, child_schedule.pprint())
            if box.scheduling == SchedulingStrategy.ALAP:
                extend_schedule_new(schedule, child_schedule.reverse(), self.channels)
            else:
                extend_schedule_new(schedule, child_schedule, self.channels)

            if self._logger.isEnabledFor(logging.DEBUG):
                self._logger.debug("Updated schedule:\n%s", schedule.pprint())

        if box.scheduling == SchedulingStrategy.ALAP:
            schedule = schedule.reverse()
        self._logger.debug("'%s' resolved.", box.label)

        box.atom = schedule
        return box.atom

    def _block_neighborhood_tetris(self, schedule: Schedule, locus: set[str], neighborhood: int) -> None:
        """Add additional blocked channels to the schedule, preventing their use during the schedule.

        In the idealized computational model we assume that in a (correctly calibrated) quantum computer
        there is no effective interaction between QPU components (in the computational frame and subspace) while
        a :class:`.Wait` instruction is acting on the flux channel of the coupler connecting those components
        (i.e., the coupler is idling).
        Hence a QPU component experiences no effective evolution if Wait instructions are
        acting on its drive, flux and probe channels, and the flux channels of all its couplers.

        Of course, in reality the QPU will experience at least some decoherence during a Wait, and
        possibly some crosstalk.
        In some applications, e.g. decoherence experiments, it is essential to Wait for a specific time,
        since it's precisely the decoherence that we are interested in.

        However, if we are only interested in applying well-defined local quantum operations on specific
        loci, it is essential to shut down all unwanted dynamics by adding :class:`.Block` instructions on
        control channels of the unused and neighboring channels.
        They act like Waits (and are converted into Waits at the end of the scheduling), but are allowed
        to overlap in time, since we are only interested in blocking those channels for the duration of the
        quantum operation.

        Args:
            schedule: instruction schedule to modify
            locus: information-carrying QPU components ``schedule`` is meant to operate on (does not include couplers)
            neighborhood: How far should we block neighboring QPU components?
                Zero means just the locus qubits, one means neighboring couplers, two means *their* neighboring
                qubits etc.

        Returns:
            ``schedule``, with added :class:`.Block` instructions on all the neighbor channels,
            for the duration of the schedule

        """
        channels: list[str] = []
        if neighborhood >= 0:
            # locus channels
            channels.extend(self.get_control_channels(locus))
        if neighborhood >= 1:
            # channels of couplers neighboring the locus
            couplers = self.chip_topology.get_neighbor_couplers(locus)
            channels.extend(
                self.get_flux_channel(coupler)
                for coupler in couplers
                if coupler in self.component_channels and "flux" in self.component_channels[coupler]
            )
        if neighborhood >= 2:
            # TODO the scheduling could be improved here
            # channels of qubits neigboring the locus
            neighbors = self.chip_topology.get_neighbor_locus_components(locus)
            channels.extend(self.get_control_channels(neighbors))
        # TODO systematically allow even larger neighborhoods?

        # add the blocks
        # TODO VirtualRZ on a drive channel is no reason to block coupler fluxes!
        schedule.add_channels(channels)
        T = schedule.duration
        for ch in channels:
            # if ch already exist in the schedule, nothing happens to it
            if not schedule[ch]:
                schedule.append(ch, Block(T))

    def build_playlist(self, schedules: Iterable[Schedule], finish_schedules: bool = True) -> Playlist:
        """Build a playlist from a number of instruction schedules.

        This involves compressing the schedules so that no duplicate information
        needs to be transferred to Station Control.

        All virtual channels are dropped at this point.

        Args:
            schedules: finished instruction schedules to include in the playlist
            finish_schedules: whether to finalise the schedules before building the playlist. Should be set ``True``
                unless some process has already finalised them before calling this function.

        Returns:
            playlist containing the schedules

        Raises:
            ValueError: if the schedules contain channels with non-uniform sampling rates

        """

        # build the channel descriptions
        def _map_channel_config(props: ChannelProperties) -> IQChannelConfig | RealChannelConfig | ReadoutChannelConfig:
            if isinstance(props, ProbeChannelProperties):
                return ReadoutChannelConfig(props.sample_rate)
            if props.is_iq:
                return IQChannelConfig(props.sample_rate)
            return RealChannelConfig(props.sample_rate)

        channel_descriptions = {
            channel_name: ChannelDescription(controller_name=channel_name, channel_config=_map_channel_config(props))
            for channel_name, props in self.channels.items()
            if not props.is_virtual
        }

        pl = Playlist()
        mapped_instructions: dict[str, dict[int, Any]] = {}

        def _append_to_schedule(sc_schedule: SC_Schedule, channel_name: str, instr: Instruction) -> None:
            """Append ``instr`` to ``sc_schedule`` into the channel``channel_name``."""
            try:
                # Check if instr can be used as a dictionary key, and use it if possible.
                # 2 dataclasses can have the same hash if their fields are identical. We must
                # distinguish between different Waveform classes which may have identical fields,
                # so we use the instruction itself as a key, so that the class is checked too.
                instr_id = hash(instr)
            except TypeError:
                instr_id = instr.id
            if instr_id not in mapped_instructions.setdefault(channel_name, {}):
                mapped = _map_instruction(instr)
                idx = pl.channel_descriptions[channel_name].add_instruction(mapped)
                sc_schedule.instructions.setdefault(channel_name, []).append(idx)
                mapped_instructions[channel_name][instr_id] = idx
            else:
                sc_schedule.instructions.setdefault(channel_name, []).append(
                    mapped_instructions[channel_name][instr_id]
                )

                # add the schedules in the playlist

        # NOTE that there is no implicit right-alignment or equal duration for schedules, unlike in old-style playlists!
        for schedule in schedules:
            finished_schedule = self._finish_schedule(schedule) if finish_schedules else schedule
            sc_schedule = SC_Schedule()
            for channel_name, segment in finished_schedule.items():
                if (channel := channel_descriptions.get(channel_name)) is None:
                    continue
                pl.add_channel(channel)
                prev_wait = None
                for instruction in segment:
                    # convert all NONSOLID instructions into Waits
                    if finish_schedules and (isinstance(instruction, NONSOLID) or isinstance(instruction, Wait)):
                        if instruction.duration > 0:
                            if prev_wait:  # if the previous instruction was a Wait, combine durations
                                prev_wait = Wait(prev_wait.duration + instruction.duration)
                            else:
                                prev_wait = Wait(instruction.duration)
                    else:
                        if prev_wait:  # if there's a prev_wait not yet added to schedule, place it before instruction
                            instructions_to_add = [prev_wait, instruction]
                            prev_wait = None
                        else:
                            instructions_to_add = [instruction]
                        for instr in instructions_to_add:
                            _append_to_schedule(sc_schedule, channel_name, instr)
                if prev_wait:
                    _append_to_schedule(sc_schedule, channel_name, prev_wait)
            pl.segments.append(sc_schedule)
        return pl

    def _set_gate_implementation_shortcut(self, op_name: str) -> None:
        """Create shortcut for `self.get_implementation(<op_name>, ...)` as `self.<op_name>(...)`.

        If there is a name collision with another attribute in ``self``, the shortcut method won't be added and
        a warning is raised.
        """

        def _shortcut_mthd(
            self,
            locus: Iterable[str],
            impl_name: str | None = None,
            *,
            use_priority_order: bool = False,
            strict_locus: bool = False,
            priority_calibration: OILCalibrationData | None = None,
        ) -> GateImplementation:
            return self.get_implementation(
                op_name,
                locus,
                impl_name=impl_name,
                use_priority_order=use_priority_order,
                strict_locus=strict_locus,
                priority_calibration=priority_calibration,
            )

        if not hasattr(self, op_name):
            setattr(self, op_name, MethodType(_shortcut_mthd, self))
        else:
            warning_msg = (
                f"Shortcut method ``ScheduleBuilder.{op_name}`` for "
                f'``ScheduleBuilder.get_implementation("{op_name}", ...)`` was not added as there is already'
                f"a class attribute ``ScheduleBuilder.{op_name}``."
            )
            self._logger.warning(warning_msg)


def _map_instruction(inst: Instruction) -> sc_instructions.Instruction:
    """TODO only necessary until SC has been updated to use the iqm.pulse Instruction class."""
    operation: Any

    def _map_acquisition(acq: AcquisitionMethod) -> sc_instructions.AcquisitionMethod:
        if isinstance(acq, ThresholdStateDiscrimination):
            return sc_instructions.ThresholdStateDiscrimination(
                label=acq.label,
                delay_samples=acq.delay_samples,
                weights=_map_instruction(acq.weights).operation,
                threshold=acq.threshold,
                feedback_signal_label=acq.feedback_signal_label,
            )
        if isinstance(acq, ComplexIntegration):
            return sc_instructions.ComplexIntegration(
                label=acq.label, delay_samples=acq.delay_samples, weights=_map_instruction(acq.weights).operation
            )
        if isinstance(acq, TimeTrace):
            return sc_instructions.TimeTrace(
                label=acq.label, delay_samples=acq.delay_samples, duration_samples=acq.duration_samples
            )
        raise ValueError(f"Unknown AcquisitionMethod {acq}")

    if isinstance(inst, Wait):
        operation = sc_instructions.Wait()
    elif isinstance(inst, VirtualRZ):
        operation = sc_instructions.VirtualRZ(inst.phase_increment)
    elif isinstance(inst, RealPulse):
        operation = sc_instructions.RealPulse(
            wave=to_canonical(inst.wave),
            scale=inst.scale,
        )
    elif isinstance(inst, IQPulse):
        operation = sc_instructions.IQPulse(
            wave_i=to_canonical(inst.wave_i),
            wave_q=to_canonical(inst.wave_q),
            scale_i=inst.scale_i,
            scale_q=inst.scale_q,
            phase=inst.phase,
            modulation_frequency=inst.modulation_frequency,
            phase_increment=inst.phase_increment,
        )
    elif isinstance(inst, ConditionalInstruction):
        if len(inst.outcomes) != 2:
            raise ValueError("ConditionalInstruction requires exactly two outcomes.")
        operation = sc_instructions.ConditionalInstruction(
            condition=inst.condition,
            if_true=_map_instruction(inst.outcomes[1]),
            if_false=_map_instruction(inst.outcomes[0]),
        )
    elif isinstance(inst, MultiplexedIQPulse):
        sc_entries = tuple((_map_instruction(p), d) for p, d in inst.entries)
        operation = sc_instructions.MultiplexedIQPulse(sc_entries)
    elif isinstance(inst, ReadoutTrigger):
        sc_acquisitions = tuple(_map_acquisition(a) for a in inst.acquisitions)
        operation = sc_instructions.ReadoutTrigger(
            probe_pulse=_map_instruction(inst.probe_pulse),
            acquisitions=sc_acquisitions,
        )
    else:
        raise ValueError(f"{inst} not supported.")
    return sc_instructions.Instruction(duration_samples=int(inst.duration), operation=operation)



#  ********************************************************************************
#
# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base utility functions with no dependencies on other iqm.pulse modules."""

from __future__ import annotations

from collections.abc import Sized
import copy
from typing import Any

import numpy as np


def merge_dicts(A: dict, B: dict, path=(), merge_nones: bool = True) -> dict:
    """Merge two dictionaries recursively, leaving the originals unchanged.

    Args:
        A: dictionary
        B: another dictionary
        merge_nones: whether to also merge ``None`` and empty ``Sized`` values from B to A.

    Returns:
        copy of A, with the contents of B merged in (and taking precedence) recursively

    """

    def is_not_empty(val: Any) -> bool:
        if val is None:
            return False
        if isinstance(val, Sized) and len(val) == 0:
            return False
        return True

    # A and B must be left intact, make a shallow copy
    A = copy.copy(A)
    for key, vb in B.items():
        if (va := A.get(key)) is not None:
            new_path = (*path, key)
            if isinstance(va, dict):
                if isinstance(vb, dict):
                    # replace the dict with the shallow copy returned by merge_dicts
                    A[key] = merge_dicts(va, vb, new_path, merge_nones=merge_nones)
                    continue
                raise ValueError(f"Merging dict with scalar: {'.'.join(new_path)}")
            if isinstance(vb, dict):
                raise ValueError(f"Merging scalar with dict: {'.'.join(new_path)}")
        # scalar overrides scalar, or a new key is inserted
        if merge_nones or is_not_empty(vb):
            A[key] = vb
    return A  # return the shallow copy


def _dicts_differ(a: Any, b: Any) -> bool:
    if isinstance(a, dict) and isinstance(b, dict):
        if a.keys() != b.keys():
            return True
        return any(_dicts_differ(a[key], b[key]) for key in a)
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        return not np.array_equal(a, b)
    return a != b
